from . import polarize as pol
from . import smooth as ks
# import kernel_smoothing_functions as ks
import numpy as np
import copy
import pandas as pd

def check_both_none(thing1,thing2):
    '''
    Utility function to check if two things are both None.
    Returns True if both are None, False otherwise.
    '''
    if thing1 is None and thing2 is None:
        return True
    return False    

def check_differ_none(thing1,thing2):
    '''
    Utility function to check if only one of two things is None.
    Returns True if one is None and the other is not, False otherwise.
    '''
    if thing1 is None and thing2 is not None:
        return True
    if thing1 is not None and thing2 is None:
        return True
    return False
    

def compare_DiemTypes(dTest,dSolution):

    testPassed = True
    AttributesFailedList = []

    # check that the attributes of the test DiemType match those of the solution DiemType
    # we assume that both dTest and dSolution are DiemType objects
    # we also assume that they have the same number of chromosomes and individuals
    # we do not check that here, but we could add that if needed
    # we also assume that the order of individuals and chromosomes is the same in both objects
    # if that is not the case, we could add code to reorder them to match
    # but for now we just assume they are the same
   

    # dTest could have additional attributes than dSolution.
    # we only check attributes that (should) be shared between them
    # we check list attributes, non-list attributes, and float attributes separately, then we check if the contig matrices are the same

    print("Now checking list attributes")
    for key in ['DMBC','PolByChr','DIByChr','SupportByChr','posByChr','MapBC','chrPloidies','initialPolByChr']:
        print('checking',key)

        if check_both_none(dSolution.__dict__[key], dTest.__dict__[key]):
            #print("Both are None, so continuing")
            continue
        if check_differ_none(dSolution.__dict__[key], dTest.__dict__[key]):
            print("Attribute ",key," differs (one is None, the other is not)")
            testPassed = False
            AttributesFailedList.append(key)
            continue
        
        # in all other cases, just compare the arrays within the lists
        for idx in range(len(dSolution.DMBC)):
            arr1 = dSolution.__dict__[key][idx]
            arr2 = dTest.__dict__[key][idx]
            if not np.array_equal(arr1,arr2):
                print("Attribute ",key," differs for chromosome ",idx)
                testPassed = False
                AttributesFailedList.append(key)
    
    print("Now checking non-list attributes")
    for key in ['indNames','chrNames','chrLengths','HIs']:
        #this code is to handle the case where HIs have not been computed, and are thus None

        if check_both_none(dSolution.__dict__[key], dTest.__dict__[key]):
            #print("Both are None, so continuing")
            continue
        if check_differ_none(dSolution.__dict__[key], dTest.__dict__[key]):
            print("Attribute ",key," differs (one is None, the other is not)")
            testPassed = False
            AttributesFailedList.append(key)
            continue

        # in all other cases, just compare the arrays
        arr1 = dSolution.__dict__[key]
        arr2 = dTest.__dict__[key]
        if not np.array_equal(arr1,arr2):
            print("Attribute ",key," differs")
            testPassed = False
            AttributesFailedList.append(key)


    print("Now checking float attributes")
    for key in ['threshold','smoothScale']:
        if check_both_none(dSolution.__dict__[key], dTest.__dict__[key]):
            #print("Both are None, so continuing")
            continue
        if check_differ_none(dSolution.__dict__[key], dTest.__dict__[key]):
            print("Attribute ",key," differs (one is None, the other is not)")
            testPassed = False
            AttributesFailedList.append(key)
            continue

        # in all other cases, just compare the values
        if dSolution.__dict__[key] != dTest.__dict__[key]:
            print("Attribute ",key," differs")
            testPassed = False
            AttributesFailedList.append(key)

    # Compare relativeRecRateDict attribute
    print("Now checking relativeRecRateDict attribute")
    key = 'relativeRecRateDict'
    # Check if both have the attribute
    has_attr_solution = hasattr(dSolution, key)
    has_attr_test = hasattr(dTest, key)
    if not has_attr_solution and not has_attr_test:
        pass  # Both do not have the attribute, skip
    elif has_attr_solution != has_attr_test:
        print(f"Attribute {key} differs (one object has it, the other does not)")
        testPassed = False
        AttributesFailedList.append(key)
    else:
        # Both have the attribute, compare values
        val1 = getattr(dSolution, key)
        val2 = getattr(dTest, key)
        if check_both_none(val1, val2):
            pass
        elif check_differ_none(val1, val2):
            print(f"Attribute {key} differs (one is None, the other is not)")
            testPassed = False
            AttributesFailedList.append(key)
        elif val1 != val2:
            print(f"Attribute {key} differs")
            testPassed = False
            AttributesFailedList.append(key)

    print('now checking individual exclusions')
    key = 'indExclusions'
    if check_both_none(dSolution.__dict__[key], dTest.__dict__[key]):
        #print("Both are None, so continuing")
        pass
    elif check_differ_none(dSolution.__dict__[key], dTest.__dict__[key]):
        print("Attribute ",key," differs (one is None, the other is not)")
        testPassed = False
        AttributesFailedList.append(key)
    else:
        same = np.array_equal(np.sort(dSolution.__dict__[key]), np.sort(dTest.__dict__[key]))
        if not same:
            print("Attribute ",key," differs")
            testPassed = False
            AttributesFailedList.append(key)
    
    print('now checking site exclusions')
    key = 'siteExclusionsByChr'
    if check_both_none(dSolution.__dict__[key], dTest.__dict__[key]):
        #print("Both are None, so continuing")
        pass
    elif check_differ_none(dSolution.__dict__[key], dTest.__dict__[key]):
        print("Attribute ",key," differs (one is None, the other is not)")
        testPassed = False
        AttributesFailedList.append(key)
    else:
        for idx in range(len(dSolution.__dict__['chrNames'])):
            arr1 = dSolution.__dict__[key][idx]
            arr2 = dTest.__dict__[key][idx]
            if check_both_none(arr1, arr2):
                #print("Both are None, so continuing")
                continue
            if check_differ_none(arr1, arr2):
                print("Attribute ",key," differs for chromosome ",dSolution.__dict__['chrNames'][idx]," (one is None, the other is not)")
                testPassed = False
                AttributesFailedList.append(key)
                break
            if not np.array_equal(np.sort(arr1), np.sort(arr2)):
                print("Attribute ",key," differs for chromosome ",dSolution.__dict__['chrNames'][idx])
                testPassed = False
                AttributesFailedList.append(key)
                break


    if testPassed == False:
        print('previous attributes failed, so skipping contigMatrix check')
    else:
        print("now checking contigMatrix attribute")
        key = 'contigMatrix'
        if check_both_none(dSolution.__dict__[key], dTest.__dict__[key]):
            #print("Both are None, so continuing")
            pass
        elif check_differ_none(dSolution.__dict__[key], dTest.__dict__[key]):
            print("Attribute ",key," differs (one is None, the other is not)")
            testPassed = False
            AttributesFailedList.append(key)
        else:
            if not compare_contig_matrices(dSolution.__dict__[key],dTest.__dict__[key]):
                print("Attribute ",key," differs (neither is None) ")
                testPassed = False
                AttributesFailedList.append(key)
            


    if testPassed:
        print("All tests passed!")
    else:
        AttributesFailedList = list(set(AttributesFailedList))
        print("The following attributes differed: ",AttributesFailedList)   


def compare_intervals(iv1,iv2):
    # compares two Interval objects
    # returns True if they are the same, False otherwise
    attrs = ['chrName','indName','idxl','idxr','l','r','state']
    for attr in attrs:
        if getattr(iv1,attr) != getattr(iv2,attr):
            print("Attribute ",attr," differs: ",getattr(iv1,attr)," vs ",getattr(iv2,attr))
            return False
    return True

def compare_interval_lists(list1,list2):
    # compares two lists of Interval objects
    # returns True if they are the same, False otherwise
    if len(list1) != len(list2):
        print("Lists differ in length: ",len(list1)," vs ",len(list2))
        return False
    for iv1,iv2 in zip(list1,list2):
        if not compare_intervals(iv1,iv2):
            return False
    return True


def compare_contig(c1,c2):
    # compares two Contig objects
    # returns True if they are the same, False otherwise

    if c1.chrName != c2.chrName or c1.indName != c2.indName:
        print("Contigs differ in chrName or indName: ",c1.chrName,c1.indName," vs ",c2.chrName,c2.indName)
        return False
    if not compare_interval_lists(c1.intervals,c2.intervals):
        print("Contigs differ in intervals")
        return False
    return True


def compare_contig_matrices(cm1,cm2):
    # compares the contigMatrix of two different contig matrices objects. It is an array with rows = chromosomes and columns = individuals.

    # returns True if they are the same, False otherwise. Could return more information about where they differ if needed.
    # print(cm1)
    # print(cm2)
    if check_both_none(cm1,cm2):
        #print("Both contigMatrices are None")
        return True
    elif check_differ_none(cm1,cm2):
        print("contigMatrix differs (one is None, the other is not)")
        return False
    else:
        nChrs = len(cm1)
        nInds = len(cm1[0])
        for idxChr in range(nChrs):
            for idxInd in range(nInds):
                c1 = cm1[idxChr][idxInd]
                c2 = cm2[idxChr][idxInd]
                if not compare_contig(c1,c2):
                    print("Contigs differ at chr index ",idxChr," and ind index ",idxInd)
                    return False
        return True


#likely need to adjust what needs being imported, 
#for local imports, check which functions are being used and whether there are better functions defined elsewhere

import numpy as np
import os
import pandas as pd
# np.set_printoptions(legacy = '1.25')

#import pandas as pd
#import csv
#import time
#import os
#import multiprocessing
# from collections import Counter
# import pickle

# from . import polarize as pol
# from . import smooth as ks

# left and right indices are inclusive. So interval is [l,r] inclusive of both ends.
def get_intervals(chrName, indName, statesList, posList, includeSingle = True):
    '''
    Function to get intervals for a given contig from states and positions. 
    single site intervals can be included or excluded. By default they are included.

    Args:

        chrName (str): Chromosome name.
        indName (str): Individual name.
        statesList (list): List of states at each site for this individual and chromosome.
        posList (list): List of positions.
        mapPosList (list): List of map positions.

    Returns:
        list: List of intervals.
    '''

    lidx = 0
    ridx = 0
    ivls = []

    if len(statesList) == 0:
        return ivls
    
    while ridx <= len(statesList)-1:
        currentState = statesList[lidx]

        if ridx == len(statesList)-1:
            l = posList[lidx]
            r = posList[ridx]
            iv = Interval(chrName, indName, lidx, ridx, l, r, currentState)

            if includeSingle == True:
                ivls.append(iv)
            else:
                if r-l>0: 
                    ivls.append(iv)
            break

        if statesList[ridx+1] == currentState:
            ridx += 1
        else:
            l = posList[lidx]
            r = posList[ridx]

            iv = Interval(chrName, indName, lidx, ridx, l, r, currentState)

            if includeSingle == True:
                ivls.append(iv)
            else:
                if r-l>0: 
                    ivls.append(iv)

            ridx +=1
            lidx=ridx
            
    return ivls




# interval should also add info about (any) sites supporting the interval in addition to the left and right sites defining it
# this should be done in a way to look for gene conversion events later on
class Interval:
    '''
    Represents a genomic interval for a specific individual and chromosome.

    Args:

        chrName (str): Chromosome name.
        indName (str): Individual name.
        idxl (int): Left index (inclusive).
        idxr (int): Right index (inclusive). So slice of state matrix would be [idxl:idxr+1]
        l (float): Left position (physical).
        r (float): Right position (physical).
        state (int): State of the interval.

    :ivar str chrName: Chromosome name.
    :ivar str indName: Individual name.
    :ivar int idxl: Left index (inclusive).
    :ivar int idxr: Right index (inclusive). So slice of state matrix would be [idxl:idxr+1]
    :ivar float l: Left position (physical).
    :ivar float r: Right position (physical).
    :ivar int state: State of the interval.

    '''

    def __init__(self,chrName,indName,idxl,idxr,l,r,state):

        self.chrName = chrName
        self.indName = indName
        self.idxl = idxl
        self.idxr = idxr
        self.l = l
        self.r = r
        self.state = state



    def info(self):
        print(f"chr = {self.chrName}, ind = {self.indName}, idxl = {self.idxl}, idxr = {self.idxr}, l = {self.l}, r = {self.r}, state = {self.state}")

    def span(self):
        return self.r - self.l
    
    def mapSpan(self,chrLength):
        return (self.r - self.l)/chrLength # in on a 'unit map length' scale, i.e. 0 to 1
    

    
#the chromosome class contains the haplotype structure of a single chromosome.
# the individual and chromosome are indexedIt contains 
# for a single individual and single chromosome. Maybe 'individualChromsome' would be a better name?
# the individual is simply a list of the intervals (as defined above)
class Contig:
    
    '''
    Represents a contiguous sequence of genomic intervals for a specific individual and chromosome.
    
    Args:
        chrName (str): Chromosome name.
        indName (str): Individual name.
        intervalList (list): List of Interval objects.

    :ivar str chr: Chromosome name.
    :ivar str ind: Individual name.
    :ivar int num_intervals: Number of intervals.
    :ivar list intervals: List of Interval objects.

    '''

    
    
    # individual is a list of intervals pertaining to a single chromosome

    def __init__(self,chrName=None,indName=None,intervalList=None):
  
        self.intervals = intervalList
        if self.intervals is None:
            self.num_intervals = 0
        else:
            self.num_intervals = len(self.intervals)
        self.indName = indName
        self.chrName = chrName

        # self.getZeroIntervals()
        # self.getOneIntervals()
        # self.getTwoIntervals()
        # self.getThreeIntervals()
        
    

    def printIntervals(self,lim=10):
        print("formatting is as follows [leftPosition,rightPosition,state]")
        if lim is None:
            print([[x.l,x.r,x.state] for x in self.intervals]) 
        else:
            print([[x.l,x.r,x.state] for x in self.intervals[0:min(self.num_intervals,lim)]])


    def get_my_intervals_of_state(self,state):
        ivs = []
        for x in self.intervals:
            if x.state == state:
                ivs.append(x)
        return ivs


def pack_interval(interval):
    '''
    Packs an Interval object into a dictionary.
    Args:
        interval (Interval): Interval object to be packed.
    Returns:
        dict: Dictionary representation of the Interval object.
    '''
    return interval.__dict__

def unpack_interval(d):
    '''
    Unpacks a dictionary into an Interval object.
    Args:
        d (dict): Dictionary representation of an Interval object.
    Returns:
        Interval: Unpacked Interval object.
    '''
    blankArgs = [None for _ in range(7)]
    i = Interval(*blankArgs)
    i.__dict__.update(d)
    return i

def pack_intervalList(ivl):
    '''
    Packs a list of Interval objects into a list of dictionaries.
    Args:
        ivl (list): List of Interval objects to be packed.
    Returns:
        list: List of dictionary representations of the Interval objects.
    '''
    return [pack_interval(iv) for iv in ivl]

def unpack_intervalList(dlist):
    '''
    Unpacks a list of dictionaries into a list of Interval objects.
    Args:
        dlist (list): List of dictionary representations of Interval objects.
    Returns:
        list: List of unpacked Interval objects.
    '''
    return [unpack_interval(d) for d in dlist]

def pack_contig(contig):
    '''
    Packs a Contig object into a dictionary. This requires 'packing' the list of Interval objects as well.
    Args:
        contig (Contig): Contig object to be packed.
    Returns:
        dict: Dictionary representation of the Contig object.
    '''
    d = contig.__dict__.copy()
    d['intervals'] = pack_intervalList(contig.intervals)
    return d

def unpack_contig(d):
    contig = Contig()

    contig.__dict__.update(d)
    contig.intervals = unpack_intervalList(d['intervals'])
    return contig

def pack_contig_matrix(cArr):
    '''
    Packs a Matrix of Contig objects into a matrix of dictionaries.
    Matrix is (num_chromosomes, num_individuals) in sort order of the diemtype parent object.

    Args:
        cArr (np.array dtype=object): Matrix of Contig objects to be packed.
    Returns:
        list: List of dictionary representations of the Contig objects.
    '''
    
    return [[pack_contig(c) for c in row] for row in cArr]

def unpack_contig_matrix(dArr):
    '''
    Unpacks a Matrix of dictionaries into a matrix of Contig objects.
    Matrix is (num_chromosomes, num_individuals) in sort order of the diemtype parent object.

    Args:
        dArr (list): List of dictionary representations of Contig objects.
    Returns:
        np.array dtype=object: Matrix of unpacked Contig objects.
    '''
    return np.array([[unpack_contig(d) for d in row] for row in dArr], dtype=object)






def build_contig_matrix(diemType,includeSingle = True):
    '''
    Creates a matrix of Contig objects from a DiemType object.
    Matrix is (num_chromosomes, num_individuals) in sort order of the diemtype parent object.

    Args:
        diemType (DiemType): DiemType object from which to create the Contig matrix.
    Returns:
        np.array dtype=object: Matrix of Contig objects.
    '''
    nChrs = len(diemType.chrNames)
    nInds = len(diemType.indNames)

    cArr = np.empty((nChrs, nInds), dtype=object)

    for cIdx in range(nChrs):
        chrName = diemType.chrNames[cIdx]
        for indIdx in range(nInds):
            indName = diemType.indNames[indIdx]
            statesList = diemType.DMBC[cIdx][indIdx]
            posList = diemType.posByChr[cIdx]

            ivl = get_intervals(chrName, indName, statesList, posList, includeSingle=includeSingle)
            contig = Contig(chrName, indName, ivl)
            cArr[cIdx, indIdx] = contig

    return cArr


def export_contigs_to_ind_bed_files(diemType, outputDir):
    '''
    Exports contig intervals to BED files for each individual.

    Args:
        diemType (DiemType): DiemType object containing contig data.
        outputDir (str): Directory where BED files will be saved.
    '''

    nChrs = len(diemType.chrNames)
    nInds = len(diemType.indNames)

    # Ensure output directory exists
    os.makedirs(outputDir, exist_ok=True)

    for indIdx in range(nInds):

        indName = diemType.indNames[indIdx]
        bedFilePath = os.path.join(outputDir, f"{indName}_contigs.bed")
        data = []

        with open(bedFilePath, 'w') as bedFile:
            for cIdx in range(nChrs):
                thisContig = diemType.contigMatrix[cIdx][indIdx]
            
                for ivl in thisContig.intervals:
                    data.append([ivl.chrName, ivl.l-1, ivl.r-1, ivl.state])  # BED format is 0-based start, 1-based end
        dfInd = pd.DataFrame(data, columns=["chrom", "start", "end", "state"])
        dfInd.to_csv(bedFilePath, sep="\t", header=True, index=False)


# This version does 'U' for state 0, and '0','1','2' for states 1,2,3
# def export_contigs_to_ind_bed_files(diemType, outputDir):
#     '''
#     Exports contig intervals to BED files for each individual.

#     Args:
#         diemType (DiemType): DiemType object containing contig data.
#         outputDir (str): Directory where BED files will be saved.
#     '''

#     nChrs = len(diemType.chrNames)
#     nInds = len(diemType.indNames)

#     # Ensure output directory exists
#     os.makedirs(outputDir, exist_ok=True)

#     for indIdx in range(nInds):

#         indName = diemType.indNames[indIdx]
#         bedFilePath = os.path.join(outputDir, f"{indName}_contigs.bed")
#         data = []

#         with open(bedFilePath, 'w') as bedFile:
#             for cIdx in range(nChrs):
#                 thisContig = diemType.contigMatrix[cIdx][indIdx]
#                 # change diemtype state vaules of 0,1,2,3 to 'U', '0', '1', '2'
#                 for ivl in thisContig.intervals:
#                     state = ivl.state
#                     if state == 0:
#                         state = 'U'
#                     else:
#                         state = str(state - 1)  # Convert 1,2,3 to '0','1','2'
#                     data.append([ivl.chrName, ivl.l-1, ivl.r-1, state])  # BED format is 0-based start, 1-based end
#         dfInd = pd.DataFrame(data, columns=["chrom", "start", "end", "state"])
#         dfInd.to_csv(bedFilePath, sep="\t", header=True, index=False)


import numpy as np
#np.set_printoptions(legacy = '1.25')

import pandas as pd
import csv
import time
import os
import multiprocessing
from collections import Counter
import pickle


from . import polarize as pol 
from . import smooth as ks


import matplotlib.pyplot as plt
from matplotlib import colors


# copied from https://github.com/Studenecivb/CarpePy/blob/main/carpepy/core_functions.py#L374
# to match the colors across implementations
diemColors4 = [
    'white',
    colors.to_hex((128/255, 0, 128/255)),  # RGBColor[128/255, 0, 128/255] - Purple
    colors.to_hex((255/255, 229/255, 0)),  # RGBColor[255/255, 229/255, 0] - Yellow
    colors.to_hex((0, 128/255, 128/255))   # RGBColor[0, 128/255, 128/255] - Teal
]
diemColors5 = [
    'LightGray',
    'white',
    colors.to_hex((128/255, 0, 128/255)),  # RGBColor[128/255, 0, 128/255] - Purple
    colors.to_hex((255/255, 229/255, 0)),  # RGBColor[255/255, 229/255, 0] - Yellow
    colors.to_hex((0, 128/255, 128/255))   # RGBColor[0, 128/255, 128/255] - Teal
]
bounds5 = [-1.5,-0.5,.5,1.5,2.5,3.5]


'''
Some utility functions for working with the data and results
'''


def plot_painting(diemMatrix,names=None,figSize = None,outputPath=None):
    '''
     Plot a painting of a diemMatrix. that is, an element of diemtype.DMBC, i.e., a diem matrix for a single chromosome.
     
    Args:
        diemMatrix: A diem matrix for a single chromosome (element of diemtype.DMBC)
        names: Optional list of names for each individual (to label y-axis)
        figSize: Optional tuple specifying figure size (width, height)
        outputPath: Optional path to save the figure. Extension determines format (e.g., .png, .pdf)
        names: Optional list of names for each individual (to label y-axis)
     '''
    # for a single chromosome in the 'stateMatrixByChromsome'

    if figSize is None:
        figSize = (12,4)
    mycmap = colors.ListedColormap( diemColors4)
    bounds = [-.5,.5,1.5,2.5,3.5]
    norm = colors.BoundaryNorm(bounds,mycmap.N)
    
    fig, ax = plt.subplots(figsize=figSize)
    ax.pcolormesh(diemMatrix, cmap = mycmap, norm=norm)
    
    if names is not None:
        # Set y-tick positions at the center of each row
        ax.set_yticks(np.arange(len(names)) + 0.5)
        ax.set_yticklabels(names)
    else:
        ax.set_yticks([])
        ax.set_yticklabels([])
    
    if outputPath is not None:
        plt.savefig(outputPath,bbox_inches='tight')


    plt.show()


def plot_painting_with_positions(diemMatrix, positions,markerWidth=25,figSize = None,outputPath=None,names=None):
    '''
    Plot a painting of a diemMatrix with proper spacing based on physical positions,
    showing blank spaces between markers.
    
    Args:
        diemMatrix: A diem matrix for a single chromosome (element of diemtype.DMBC)
        positions: Array of physical positions for each marker (element of diemtype.posByChr)
        markerWidth: Half-width of each marker's visual representation (in base pairs)
        figSize: Optional tuple specifying figure size (width, height)
        outputPath: Optional path to save the figure. Extension determines format (e.g., .png, .pdf)
        names: Optional list of names for each individual (to label y-axis
    '''
    # for a single chromosome in the 'stateMatrixByChromsome'

    if figSize is None:
        figSize = (12,4)
    mycmap = colors.ListedColormap(diemColors4)
    bounds = [-.5,.5,1.5,2.5,3.5]
    norm = colors.BoundaryNorm(bounds,mycmap.N)
    
    fig, ax = plt.subplots(figsize=figSize)
    
    # Each marker gets a narrow bin (2 bp wide: 1 bp left and right of position)
    # Gaps between markers will show as blank space
    x_edges = np.zeros(len(positions) + 1)
    
    if len(positions) == 1:
        # Handle single position case
        x_edges[0] = positions[0] - 1
        x_edges[1] = positions[0] + 1
    else:
        # Create edges for each marker spanning exactly 1 bp on each side
        # This will leave gaps where markers are far apart
        for i in range(len(positions)):
            x_edges[i] = positions[i] - 1
        x_edges[-1] = positions[-1] + 1
    
    # Individual boundaries (one more than number of individuals)
    y_edges = np.arange(diemMatrix.shape[0] + 1)
    
    # Plot each marker individually to show gaps
    for i in range(len(positions)):
        # Create a small mesh for this single marker
        x_marker = [positions[i] - markerWidth, positions[i] + markerWidth]
        X_marker, Y_marker = np.meshgrid(x_marker, y_edges)
        
        # Extract data for this marker (column i of the matrix)
        marker_data = diemMatrix[:, i:i+1]  # Keep as 2D array
        
        # Plot this marker
        ax.pcolormesh(X_marker, Y_marker, marker_data, cmap=mycmap, norm=norm)
    

    # Add y-axis labels if names are provided
    if names is not None:
        ax.set_yticks(np.arange(len(names)) + 0.5)
        ax.set_yticklabels(names)
    else:
        ax.set_yticks([])
        ax.set_yticklabels([])

    ax.set_xlabel('Position')
    
    # Set x-axis limits to show the full range
    ax.set_xlim(positions[0] - (positions[-1] - positions[0]) * 0.05, 
                positions[-1] + (positions[-1] - positions[0]) * 0.05)
    
    # Format x-axis to show positions in a readable way
    ax.ticklabel_format(style='scientific', axis='x', scilimits=(0,0))
    
    if outputPath is not None:
        plt.savefig(outputPath,bbox_inches='tight')

    #plt.tight_layout()
    plt.show()

def characterize_markers(dt):
    '''
    Given a diemtype instance, dt, return a dataframe with information about the marker configurations
    including their counts, DI values, number of individuals differing from ideal marker, and number of individuals with missing data
    '''
    #dt is a diemtype instance
    dmat_cat = np.hstack(dt.DMBC)
    di_cat = np.hstack(dt.DIByChr)

    markerMatrix = dmat_cat.transpose()

    dictDI = {}
    dictCount = {}
    
    for idx,marker in enumerate(markerMatrix):
        config = tuple(marker)
        
        if not config in dictDI:
            dictDI[config] = di_cat[idx]

        if config in dictCount:
            dictCount[config]+=1
        else:
            dictCount[config] = 1

    configList = list( dictDI.keys())
    diList = [dictDI[k] for k in configList]
    countList = [dictCount[k] for k in configList]

    df = pd.DataFrame({'configuration':configList,'count':countList,'DI':diList})
    df.sort_values(by = 'DI',ascending=False,inplace=True,ignore_index=True)

    # adding in additional info...
    idealMarker = df.iloc[0]['configuration']
    diffsFromIdealList = []
    sitesMissingList = []
    for marker in df['configuration']:
        kdiffs = np.sum(np.array(idealMarker)!=np.array(marker))
        diffsFromIdealList.append(kdiffs)
        kmissing = np.sum(np.array(marker)==0)
        sitesMissingList.append(kmissing)
    df['diffs_from_ideal'] = diffsFromIdealList
    df['individuals_missing'] = sitesMissingList
    return df
    

def check_coverage_after_thresholding(dt,threshold):
    '''
    Given a diemtype instance, dt, and a threshold value, return the proportion of sites retained after thresholding
    '''
    #dt is a diemtype instance
    propRetainedByChr = []
    for idx,di in enumerate(dt.DIByChr):
        propRetainedByChr.append( sum(di>=threshold)/len(di))

    return propRetainedByChr

def count_site_differences(dmbc1,dmbc2):
    '''
    Given two lists of diem matrices (that is, by chromosome), count the number of sites that differ between them
    '''
    if len(dmbc1) != len(dmbc2):
        raise ValueError("count_site_differences: diem matrices must have the same number of chromosomes")

    kdiffs = 0
    for chr1, chr2 in zip(dmbc1, dmbc2):
        if chr1.shape != chr2.shape:
            raise ValueError("count_site_differences: diem matrices must have the same shape")
        kdiffs += np.sum(chr1 != chr2)

    return kdiffs

# some functions below used to make dataframes for investigating the effect of thresholding and stuff like that
# these are not currently incorporated into the main workflow but were used, e.g., in making the eseb poster
# and they should be incorporated properly in the data analysis procedure. 

# DS: We used this for Zia's project for exploratory analysis, but I don't think it is needed or wanted for the main workflow
# def remove_missing(stateMatrixByChr, positionsByChr, polarityByChr, DIByChr, mapPosByChr):

#     smbc = []
#     posbc = []
#     polbc = []
#     dibc = []
#     mpbc = []

#     for idx in range(len(stateMatrixByChr)):
#         sm = stateMatrixByChr[idx]
#         pos = positionsByChr[idx]
#         pol = polarityByChr[idx]
#         DI = DIByChr[idx]
#         mapPos = mapPosByChr[idx]
        
#         trans = sm.transpose()
#         thisFilter = trans==0
#         thisFilter = np.array([sum(x)==0 for x in thisFilter])

#         trans = trans[thisFilter]
#         sm = trans.transpose()
#         pos = pos[thisFilter]
#         pol = pol[thisFilter]
#         DI = DI[thisFilter]
#         mapPos = mapPos[thisFilter]

#         smbc.append(sm)
#         posbc.append(pos)
#         polbc.append(pol)
#         dibc.append(DI)
#         mpbc.append(mapPos)

#     return smbc,posbc,polbc,dibc,mpbc
    



# below are some functions for doing basic plotting, but they are to be superceded by more advanced methods in other packages, at lesat for the advanced plotting for publication-level figures
# some functions are for statematrix in diplotype class (where to put this code?)
# some functions are for intervals in the haplolotype class (also, where to put this code?)

def plot_all_spans(d_):
    # d_ is an instance of the Diplotype class
    print("spans for all individuals. blue = hom left allele, magenta = het, red = hom right")
    spans = []
    for thisChr in d_.chromosomes:
        spans = spans + [x.mapSpan for x in thisChr.getOneIntervals()]
    n,myBins,patches = plt.hist(spans,bins=1000,histtype='step',color='b')
    
    for thisChr in d_.chromosomes:
        spans = spans + [x.mapSpan for x in thisChr.getTwoIntervals()]
    plt.hist(spans,bins=1000,histtype='step',color='m')
    
    for thisChr in d_.chromosomes:
        spans = spans + [x.mapSpan for x in thisChr.getThreeIntervals()]
    plt.hist(spans,bins=1000,histtype='step',color='r')
    
    plt.yscale("log")
    plt.xscale("log")
    plt.show()
    
    
def plot_left_spans(d_,markerBreakIdx_):
    print("spans for individuals left of barrier. blue = hom left allele, magenta = het, red = hom right")
    spans = []
    for thisChr in d_.chromosomes:
        if thisChr.ind < markerBreakIdx_:
            spans = spans + [x.mapSpan for x in thisChr.getOneIntervals()]
    n,myBins,patches = plt.hist(spans,bins=1000,histtype='step',color='b')
    
    spans = []
    for thisChr in d_.chromosomes:
        if thisChr.ind < markerBreakIdx_:
            spans = spans + [x.mapSpan for x in thisChr.getTwoIntervals()]
    plt.hist(spans,bins=1000,histtype='step',color='m')
    
    spans = []
    for thisChr in d_.chromosomes:
        if thisChr.ind < markerBreakIdx_:
            spans = spans + [x.mapSpan for x in thisChr.getThreeIntervals()]
    plt.hist(spans,bins=1000,histtype='step',color='r')
    
    plt.yscale("log")
    plt.xscale("log")
    plt.show()
    
    
def plot_right_spans(d_,markerBreakIdx_):
    print("spans for individuals right of barrier. blue = hom left allele, magenta = het, red = hom right")
    spans = []
    for thisChr in d_.chromosomes:
        if thisChr.ind >= markerBreakIdx_:
            spans = spans + [x.mapSpan for x in thisChr.getOneIntervals()]
    n,myBins,patches = plt.hist(spans,bins=1000,histtype='step',color='b')
    
    spans = []
    for thisChr in d_.chromosomes:
        if thisChr.ind >= markerBreakIdx_:
            spans = spans + [x.mapSpan for x in thisChr.getTwoIntervals()]
    plt.hist(spans,bins=1000,histtype='step',color='m')
    
    spans = []
    for thisChr in d_.chromosomes:
        if thisChr.ind >= markerBreakIdx_:
            spans = spans + [x.mapSpan for x in thisChr.getThreeIntervals()]
    plt.hist(spans,bins=1000,histtype='step',color='r')
    
    plt.yscale("log")
    plt.xscale("log")
    plt.show()



# these functions are for exploring the data a bit, used in creating my eseb poster. 
# we should consider if they are useful/ informative and whether they should be included in the final release

def get_intro_amt_and_bias(lv,rv,brakeIdx,conf):
    conf = np.array(conf)
    lconf,rconf = [conf[0:brakeIdx], conf[brakeIdx:]]
    nl = 2*sum(lconf != 0)
    nr = 2*sum(rconf != 0 )
    if nl == 0 or nr==0:
        return(np.nan,np.nan)
    introAmount = (1*sum(lconf == 2) + 2*sum(lconf == rv) + 1*sum(rconf==2) + 2*sum(rconf==lv))/(nl+nr)
    lb = (1*sum(lconf==2) + 2*sum(lconf == rv))/nl
    rb = (1*sum(rconf == 2) + 2*sum(rconf == lv))/nr
    totb = -(lb - rb)
    return introAmount,totb

# def get_thresholds(SMBC_,DIBC_):
#     sm = np.hstack(SMBC_)
#     smt = sm.transpose()
#     dis = np.concat(DIBC_)
#     bestIdx = np.argmax(dis)
#     idealMarker = smt[bestIdx]
#     kmax = len(idealMarker)
#     DIbyDiffs = np.zeros(kmax+1)
    
#     for idx,markerConfig in enumerate(smt):
        
#         kDiffs = sum(markerConfig != idealMarker)
#         thisDI = dis[idx]
#         if thisDI < DIbyDiffs[kDiffs]:
#             DIbyDiffs[kDiffs] = thisDI
#     return idealMarker, [[k,d] for k, d in enumerate(DIbyDiffs)]

#this function possibly made for Zia's project specifically, I don't recal. 
def calc_contig_his(SMBC,sexPloidy,autoPloidy,isSexChr):
    #sexPloidy is a list of the ploidy of the individuals as ordered in SMBC
    #autoPloidy is a list of the ploidy (here, always twos) as for sexPloidy, ordered by SMBC!
    #isSexChr is a list with '1' indicating sex chr and '0' indicated autosome
    HIBC = []
    for idx,sm in enumerate(SMBC):

        am = pol.hapStateIndices_to_hapMatrix(sm.transpose())
        if isSexChr[idx]==1:
            ploidyArr = sexPloidy[:,np.newaxis]
        else:
            ploidyArr = autoPloidy[:,np.newaxis]
        am = am*ploidyArr
        A4 = np.sum(am,axis=0)
        his = pol.get_hybrid_index(A4)
        HIBC.append(his)

    return np.array(HIBC)


from . import polarize as pol 
from . import smooth as ks
from . import contigs as ct
import numpy as np
import copy
import pandas as pd
import pickle
import multiprocessing as mp

# useful function for flipping the polarity of a diem matrix.
# needed so that we can read in the 'polarized' output of diem
# which stores a polarity but not the flipped matrix
# so we can just flip the matrix by the polarity on import
# can also be used to reconstruct original polarity by passing newPolarity = 0 for all markers
# and oldPolarity = the polarity stored by diem
def flip_polarity(diemMatrix,newPolarity,oldPolarity=None):
    '''
    Flip the polarity of the Diem Matrix according to the polarity array.
    diem matrix is an array with inds as rows and sites as columns
    polarity is a numpy array of shape nMarkers, with entries 0 or 1, where 1 indicates that the marker should be flipped.

    Args:
        diemMatrix (np.ndarray): The Diem Matrix to be flipped.
        newPolarity (np.ndarray): Array indicating the desired polarity for each marker (0 or 1).
        oldPolarity (np.ndarray, optional): Array indicating the current polarity for each marker (0 or 1). If oldPolarity = None, it flips the markers where newPolarity is 1.
    Returns:
        np.ndarray: The Diem Matrix with updated polarity.
    '''
    if oldPolarity is None:
        oldPolarity = np.zeros(len(newPolarity), dtype=np.int8)
    
    cols_to_flip = np.where(newPolarity != oldPolarity)[0]
    if len(cols_to_flip) == 0:
        return diemMatrix  # No changes needed

    #print('number of columns to flip:', len(cols_to_flip))
    
    arr = copy.deepcopy(diemMatrix)
    # Use a temporary value to avoid overwriting
    for col in cols_to_flip:
        col_data = arr[:, col]
        col_data[col_data == 1] = -1
        col_data[col_data == 3] = 1
        col_data[col_data == -1] = 3
    return arr
    
    # this for some reason did not work. it did not modify the array
    # but it should be faster than looping as above

    # arr[:, cols_to_flip][arr[:, cols_to_flip] == 1] = -1
    # arr[:, cols_to_flip][arr[:, cols_to_flip] == 3] = 1
    # arr[:, cols_to_flip][arr[:, cols_to_flip] == -1] = 3
    # return arr

# here, SM (State Matrix) is equivalent to DM (Diem Matrix) as it is now called
def A4_from_stateMatrix(SM,ploidies):
    I4 = np.zeros((len(SM),4))
    for idx,individual in enumerate(SM):
        numZeros = np.count_nonzero(individual==0)
        numOnes = np.count_nonzero(individual==1)
        numTwos = np.count_nonzero(individual==2)
        numThrees = np.count_nonzero(individual==3)
        I4[idx][:] = np.array([numZeros,numOnes,numTwos,numThrees])
    A4 = np.dot(np.diag(ploidies),I4)
    return A4

# defined as get_hybrid_index(A4) in diem polarization code, copied for use here
def get_hybrid_index(A4):
    #A4 is the matrix of (nHaps x states) with entries being the ploidy-adjusted counts
    nHaps = len(A4)
    HIarr = np.zeros(nHaps)
    for idx,counts in enumerate(A4):
        hiNum = counts[1]*0 + counts[2]*1 + counts[3]*2
        hiDenom = 2*(counts[1] + counts[2] + counts[3])
        hi = hiNum/hiDenom
        HIarr[idx] = hi
    return HIarr

def get_resort_order(inds1,inds2):
    """
    Get the order to resort inds2 to  match inds1
    
    Args:
        inds1 (np.ndarray): the array with the reference ordering
        inds2 (np.ndarray): the array to be reordered
    """

    newOrder = [np.argwhere(inds2 == s)[0][0] for s in inds1]
    newOrder = np.array(newOrder)

    return newOrder



# possible additions to the diemtype class:
#   re-polarize function:  takes an alternate polarity and changes state matrix. could 'de-polarize' data, as the input data is always '0' polarity
class DiemType:
    """
    Class describing the raw data for state matrices, and functions for thresholding, kernel smoothing, etc.
    
    Args:
        DMBC (List[np.ndarray]): State matrix by chromosome. For arrays, each row is an individual, each column a marker.
        indNames (np.ndarray): Names of individuals, same order as DMBC.
        chrPloidies (List[np.ndarray]): for each chromosome, the Ploidy of each individual, same order as DMBC.
        chrNames (np.ndarray): Names of chromosomes as ordered in DMBC.
        posByChr (List[np.ndarray]): Positions of markers by chromosome. 
        chrLengths (List[int]): Lengths of chromosomes. 
        exclusionsByChr (List[np.ndarray], optional): List of arrays of positions indicating which sites to exclude for each chromosome when polarizing. If None, includes all sites. If some sites are excluded, but a given chromosome has no exclusions, that list entry should be None.
        indExclusions (np.ndarray, optional): array of names of individuals to exclude when polarizing. If None, includes all individuals.

    :ivar DMBC: List[np.ndarray]. State matrix by chromosome.
    :ivar indNames: np.ndarray. Names of individuals.
    :ivar chrPloidies: List[np.ndarray]. For each chromosome, the ploidy of each individual.
    :ivar chrNames: np.ndarray. Names of chromosomes.
    :ivar posByChr: List[np.ndarray]. Positions of markers by chromosome.
    :ivar chrLengths: List[int]. Lengths of chromosomes.
    :ivar MapBC: List[np.ndarray]. Genetic map positions by chromosome, computed on initialization.
    :ivar HIs: np.ndarray. Heterozygosity indices, to be computed.
    :ivar PolByChr: List[np.ndarray]. Polymorphism matrix by chromosome, to be computed.
    :ivar initialPolByChr: List[np.ndarray]. Initial polarity by chromosome for EM start (random or test), to be computed.
    :ivar DIByChr: List[np.ndarray]. Diagnostic index by chromosome, to be computed.
    :ivar SupportByChr: List[np.ndarray]. Support values by chromosome, to be computed.
    :ivar threshold: float. Threshold value to be set.
    :ivar smoothScale: float. Scale for kernel smoothing, to be set.
    :ivar contigMatrix: np.ndarray dtype=object. Matrix of Contig objects, to be created.
    :ivar siteExclusionsByChr: List[np.ndarray]. List of arrays of positions indicating which sites to exclude for each chromosome when polarizing.
    :ivar indExclusions: np.ndarray. Array of names of individuals to exclude when polarizing.
    :ivar relativeRecRateDict: dict. Dictionary of relative recombination rates by chromosome name.

    """

    def __init__(self,DMBC, indNames, chrPloidies, chrNames, posByChr,chrLengths,exclusionsByChr=None,indExclusions=None):
            
        self.DMBC = DMBC
        self.indNames = indNames
        self.chrPloidies = chrPloidies
        self.chrNames = chrNames
        self.posByChr = posByChr
        self.chrLengths = chrLengths

        self.initMaplength()

        self.HIs = None
        self.initialPolByChr = None
        self.PolByChr = None
        self.DIByChr = None
        self.SupportByChr = None
        self.threshold = None
        self.smoothScale = None
        

        self.contigMatrix = None

        self.siteExclusionsByChr = exclusionsByChr
        self.indExclusions = indExclusions
        self.relativeRecRateDict = None
        #self.idealMarker = None
        #self.barrierBreakIdx = None

    def initMaplength(self):
        if self.posByChr is None:
            self.MapBC = None
            return None
        self.MapBC = []
        for idx in range(len(self.chrNames)):
            mapLength = self.posByChr[idx]
            mapLength = mapLength/self.chrLengths[idx]
            self.MapBC.append(mapLength)

    def add_individual_exclusions(self,filePath):
        df = pd.read_csv(filePath,header=None)
        self.indExclusions = np.array(df.iloc[:,0].tolist())

    def add_site_exclusions(self,filePath):
        self.siteExclusionsByChr = [None]*len(self.chrNames)
        skipped_chromosomes = set()
        
        with open(filePath,'r') as f:
            for line in f:
                clean_line = line.strip()
                if not clean_line:  # Skip empty lines
                    continue
                    
                chrName, start, end = clean_line.split('\t')
                start = int(start) + 1
                end = int(end) + 1

                # Check if chromosome exists in dataset
                chr_matches = np.where(self.chrNames == chrName)[0]
                if len(chr_matches) == 0:
                    skipped_chromosomes.add(chrName)
                    continue
                    
                chrIdx = chr_matches[0]
                
                if self.siteExclusionsByChr[chrIdx] is None:
                    self.siteExclusionsByChr[chrIdx] = []
                
                # Append the indices of the positions in self.posByChr[chrIdx] that are >= start and < end
                indices = np.where((self.posByChr[chrIdx] >= start) & (self.posByChr[chrIdx] < end))[0]
                self.siteExclusionsByChr[chrIdx] = self.siteExclusionsByChr[chrIdx] + indices.tolist()
        
        # Sort and deduplicate all chromosome exclusions
        for chrIdx in range(len(self.siteExclusionsByChr)):
            if self.siteExclusionsByChr[chrIdx] is not None:
                self.siteExclusionsByChr[chrIdx] = sorted(list(set(self.siteExclusionsByChr[chrIdx])))

        # Report any skipped chromosomes
        if skipped_chromosomes:
            print(f"Warning: The following chromosomes in the exclusions file were not found in the dataset and were skipped: {sorted(skipped_chromosomes)}")
        
    def add_initial_polarity(self,filePath):

        df = pd.read_csv(filePath,header=0,sep='\t')
    
        listed_names = np.unique(df['chromosome'].values.tolist())
        if set(listed_names) != set(self.chrNames):
            print("the chromosome names in the polarity file do not match those in the DiemType instance")
            return None
        
        for idx,chrName in enumerate(self.chrNames):
            
            inputPositions = df.loc[df['chromosome'] == chrName,'position'].values
            inputPositions = np.array(inputPositions,dtype=np.int32)

            if not np.array_equal(inputPositions,np.array(self.posByChr[idx])):
                print("the positions for chromosome ",chrName," in the polarity file do not match those in the DiemType instance")
                return None


            thisChrPolarity = df.loc[df['chromosome'] == chrName,'polarity'].values
            thisChrPolarity = np.array(thisChrPolarity,dtype=np.int8)
            if self.initialPolByChr is None:
                self.initialPolByChr = [None]*len(self.chrNames)
            self.initialPolByChr[idx] = thisChrPolarity
           
    def computeHIs(self,force=False):
        """
        Compute hybrid indices for each individual.
        """
        if self.PolByChr is None and force==False:
            print("data must be polarized before computing HIs. Resorting will be done automatically after polarizing, thresholding, or smoothing.")
            return None    
        else:

            A4List = []
            for idx, chr in enumerate(self.DMBC):
                ploidies = self.chrPloidies[idx]
                A4 = A4_from_stateMatrix(chr, ploidies)
                A4List.append(A4)

            A4Total = np.sum(A4List, axis=0)
            newHIs = get_hybrid_index(A4Total)

        return newHIs
            
    # this function needs to be tested with respect to contig matrix resort
    def sort(self,newHIs = None):
        """
        Sort DMBC and individuals (and their ploidies) by hybrid index.
        """
        if self.PolByChr is None:
            print("data must be polarized before sorting by HI.")
            print("Note: resorting is NOT automatic after polarizing, thresholding, or smoothing")
            print("to resort data, use the method self.sort()")
            return None
        else:
            oldOrder = np.arange(len(self.indNames))
            newHIs = self.computeHIs()    
            newOrder = np.argsort(newHIs)

            if np.all(self.HIs == newHIs) and np.all(oldOrder == newOrder):
                print("HIs already up to date, and no sorting needed")
            elif np.all(oldOrder == newOrder):
                print("HIs computed and new values updated, but ordering did not change")
            else:
                for idx,arr in enumerate(self.DMBC):
                    self.DMBC[idx] = arr[newOrder,:]
                    self.chrPloidies[idx] = self.chrPloidies[idx][newOrder]
                self.indNames = self.indNames[newOrder]
                self.HIs = newHIs[newOrder]

                if self.contigMatrix is not None:
                    for idx, arr in enumerate(self.contigMatrix):
                        self.contigMatrix[idx] = arr[newOrder,:]
                print("new HIs computer and individuals resorted by HI")

    # this function needs to be tested with respect to contig matrix resort
    def sort_as(self,d2):
        '''
        Sort the current instance (self) by the order (of individuals) of another instance (d2).
        
        Args:
            d2 (DiemType): the instance to be used for sorting
        '''
        inds1 = self.indNames
        inds2 = d2.indNames

        if np.array_equal(inds1,inds2):
            print("the datasets are already in the same order")
            return None
        
        sameSet = np.array_equal(np.sort(inds1), np.sort(inds2))
        if not sameSet:
            print("the datasets do not contain the same individuals")
            return None
        else:
            newOrder = get_resort_order(inds2,inds1)
            print(newOrder)
            for idx,arr in enumerate(self.DMBC):
                self.DMBC[idx] = arr[newOrder,:]
                self.chrPloidies[idx] = self.chrPloidies[idx][newOrder]
            self.indNames = self.indNames[newOrder]
            self.HIs = self.HIs[newOrder]


            if self.contigMatrix is not None:
                for idx, arr in enumerate(self.contigMatrix):
                    self.contigMatrix[idx] = arr[newOrder,:]

            print("dataset has been re-ordered")
            return None

    # this function needs to be tested with respect to contig matrix resort
    def sort_by(self,newOrder):
        '''
        Sort the current instance (self) by a given ordering of individuals.
        
        Args:
            newOrder (List[int] or np.ndarray): the new order of individuals, given as indices.
        '''
        # if newOrder is a list, convert to np.ndarray
        if isinstance(newOrder, list):
            newOrder = np.array(newOrder)
        # check that newOrder is a np.ndarray
        elif not isinstance(newOrder, np.ndarray):
            print("newOrder must be a list or np.ndarray")
            return None

        if len(newOrder) != len(self.indNames):
            print("the new order does not have the same length as the number of individuals")
            return None
        elif set(newOrder) != set(range(len(self.indNames))):
            print("the new order does not contain the same indices as the number of individuals")
            return None
        else:
            for idx,arr in enumerate(self.DMBC):
                self.DMBC[idx] = arr[newOrder,:]
                self.chrPloidies[idx] = self.chrPloidies[idx][newOrder]
            self.indNames = self.indNames[newOrder]
            self.HIs = self.HIs[newOrder]
            if self.contigMatrix is not None:
                for idx, arr in enumerate(self.contigMatrix):
                    self.contigMatrix[idx] = arr[newOrder,:]
            print("dataset has been re-ordered")
            return None

    def copy(self):
        '''
        Create a deep copy of the current instance.
        '''
        return copy.deepcopy(self)
    


    #note that these optional arguments are currently defined in the run_em_linear and run_em_parallel functions
    #meaning they are actually over-ridden here. Not a huge issue but could be refactored later if desired 
    def polarize(self,ncores=None,boolTestData=False,maxItt=500,epsilon=0.99999,sort_by_HI=False):

        """
        Polarize the state matrices by initializing test polarities and running the EM algorithm. Does not change self, but rather returns a polarized copy. Note that it will use the individual and site exclusions defined in self. 

        Args:
            ncores (int): number of cores to use for parallel processing. Default is None, which uses all available cores.
            boolTestData (bool): if True, initializes polarity using test data method. If False, initializes polarity randomly.
            maxItt (int, optional): Maximum number of iterations for the EM algorithm. Default is 500.
            epsilon (float, optional): Convergence threshold for the EM algorithm. Default is 0.99

        Returns:
            DiemType: A new DiemType instance with polarized data.
        """

        if ncores is None:
            ncores = mp.cpu_count()

        indExcludedIndices = None

        if self.indExclusions is not None:
            indExcludedIndices = np.where(np.isin(self.indNames,self.indExclusions))[0]


        print("convert state matrix to Marray")
        initMBC = [pol.stateMatrix_to_MArray(x) for x in self.DMBC]
        initPolBC = []

        if self.initialPolByChr is not None:
            print("using the provided initial polarity")
            initPolBC = self.initialPolByChr
            for idx, M in enumerate(initMBC):
                initMBC[idx] = pol.initialize_given_polarity(M,initPolBC[idx])
        elif boolTestData == True:
            print("initializing test polarity")
            for idx,M in enumerate(initMBC):
                thisPol,thisM = pol.initialize_test_polarity(M)
 #               print("SJEB initialized test polarity for chromosome",idx)
                initMBC[idx] = thisM
                initPolBC.append(thisPol)
        else:
            print("initializing random polarity")
            for idx,M in enumerate(initMBC):
                thisPol,thisM = pol.initialize_random_polarity(M)
                initMBC[idx] = thisM
                initPolBC.append(thisPol)
       
        if ncores == 1:
            print("running EM algorithm on a single core using diem_linear")
            MBC_out, polBC_out, DIBC_out,SupportBC_out = pol.run_em_linear(initMBC, initPolBC, self.chrPloidies,sitesExcludedByChr=self.siteExclusionsByChr, individualsExcluded=indExcludedIndices,maxItt=maxItt,epsilon=epsilon)
        else:
            print("running EM algorithm in parallel using diem_parallel with ",ncores," cores")
            MBC_out, polBC_out, DIBC_out,SupportBC_out = pol.run_em_parallel(initMBC, initPolBC, self.chrPloidies,sitesExcludedByChr=self.siteExclusionsByChr,individualsExcluded=indExcludedIndices,maxItt=maxItt,epsilon=epsilon,nCPUs=ncores)

        print("updating polarizations, DIs, Supports, initialPolarity,and state matrices")

        a = copy.deepcopy(self)
        #a = self  #this will modify the original instance, not what we want
        
        a.initialPolByChr = [x for x in initPolBC] # ensure this is a copy, not reference.
        a.PolByChr = polBC_out
        a.DIByChr = DIBC_out
        a.SupportByChr = SupportBC_out
        a.DMBC = [pol.MArray_to_stateMatrix(x) for x in MBC_out]
        
        a.HIs = a.computeHIs()
        if sort_by_HI == True:
            print("re-sorting individuals by HI")
            a.sort()

        return a
    

    def apply_threshold(self, threshold: float,sort_by_HI=False):
        """
        Apply a threshold to the diagnostic indices and update to remove sites below threshold di. Returns a copy of the modified instance.

        Args:
            threshold (float): Threshold value for diagnostic index. Sites with DI below this value will be removed.

        Returns:
            DiemType: A new DiemType instance with sites below the threshold removed.
        """

        a =copy.deepcopy(self)
        proportionRetained = []
        a.threshold = threshold
        for idx in range(len(a.DMBC)):
            myFilter = a.DIByChr[idx]>=threshold
            
            proportionRetained.append(sum(myFilter)/len(myFilter))
            trans = a.DMBC[idx].transpose()
            trans = trans[myFilter]
            a.DMBC[idx] = trans.transpose()
            a.posByChr[idx] = a.posByChr[idx][myFilter]

            a.PolByChr[idx] = a.PolByChr[idx][myFilter]
            a.DIByChr[idx] = a.DIByChr[idx][myFilter]
            a.MapBC[idx] = a.MapBC[idx][myFilter]
        
        print("after thresholding...\n")
        print("the average proportion of sites retained across chromosomes is ",np.mean(proportionRetained))
        # print the chromosome name and proportion retained for the chromosome with the smallest amount retained
        minIdx = np.argmin(proportionRetained)
        print("the chromosome with the smallest proportion of sites retained is ",a.chrNames[minIdx]," with ",proportionRetained[minIdx]," of sites retained")
        # now print the proprotion retained that is biggest
        maxIdx = np.argmax(proportionRetained)
        print("the chromosome with the largest proportion of sites retained is ",a.chrNames[maxIdx]," with ",proportionRetained[maxIdx]," of sites retained")

        print("\nthresholding done\n")
        if sort_by_HI == True:
            a.sort()
            print("new hybrid indices computed and individuals have been resorted by HI")
        else:
            a.HIs = a.computeHIs()
            print("hybrid indices have been updated")
            print("but individuals have NOT been resorted by HI.")
    
        return a
    
    def smooth(self,scale: float ,reSort=False,reSmooth=False,parallel=True):
        """
        Smooth and return a copy of the state matrices using a Laplace kernel . defaults to NOT resorting by hybrid index. This allows for direct comparison to pre-smoothed data.  May later resort using self.sort() on resulting data.

        Args:
            scale (float): Scale parameter for the Laplace kernel smoothing.
            reSort (bool, optional): If True, resorts individuals by hybrid index after smoothing. Default is False.
            reSmooth (bool, optional): If True, allows re-smoothing even if smoothing has already been done. Default is False.
            parallel (bool, optional): If True, uses parallel processing for smoothing. Default is True.

        Returns:
            DiemType: A new DiemType instance with smoothed state matrices.
        """

        if self.smoothScale is not None:
            print("smoothing has already been done with scale ",self.smoothScale)
            if reSmooth == False:
                print("In order to re-smooth the data, please set reSmooth=True. Note this will over-write the previously-used smoothing scale value")
                return None
            else:
                print("re-smoothing the data with new scale ",scale)
        else:
            print("smoothing the data with scale ",scale)

        a = copy.deepcopy(self)
        a.smoothScale = scale

        if parallel == True:
            print("using parallel smoothing")
        else:
            print("using serial smoothing")

        if self.relativeRecRateDict is not None:
            print("adjusting smoothing scale by relative recombination rates for each chromosome")    
        for idx in range(len(a.DMBC)):
            thisScale = scale
            if self.relativeRecRateDict is not None:
                chrName = a.chrNames[idx]
                if chrName in self.relativeRecRateDict:
                    thisScale = thisScale / self.relativeRecRateDict[chrName] #divide by relative rate to adjust smoothing scale
                    # print("adjusted smoothing scale for chromosome ",chrName," to ",scale)
            if parallel == False:
                thisStateMatSmoothed = ks.laplace_smooth_multiple_haplotypes(a.MapBC[idx],a.DMBC[idx],thisScale)
            else:
                thisStateMatSmoothed = ks.laplace_smooth_multiple_haplotypes_parallel(a.MapBC[idx],a.DMBC[idx],thisScale)

            a.DMBC[idx] = thisStateMatSmoothed
        if reSort == False:
            print("smoothing done, but not resorting by HI. You may call the sort() method on the resulting data if you wish to resort")
            return a
        else:
            print("smoothing done and data has been resorted by HI")
            a.sort()
            return a
        
    def create_contig_matrix(self,includeSingle = True):
        """
        Create a matrix of Contig objects from the current DiemType instance and store it in self.contigMatrix.

        Args:
            includeSingle (bool, optional): If True, includes contigs with a single marker. Default is True.
        """
        self.contigMatrix = ct.build_contig_matrix(self,includeSingle=includeSingle)
        print("contig matrix created and stored in self.contigMatrix")
        return None
    
    def get_intervals_of_state(self,state,individualSubset = None,chromosomeSubset=None):
        """
        Get intervals of a specified state for given individuals and chromosomes.

        Args:
            state (int): The state to find intervals for (0, 1, 2, or 3).
            individualSubset (List[int], optional): List of individual indices to include. If None, includes all individuals.
            chromosomeSubset (List[int], optional): List of chromosome indices to include. If None, includes all chromosomes.

        Returns:
            List[Interval]: A list of Interval objects for the specified state across the specified individuals and chromosomes.
        """
        if self.contigMatrix is None:
            print("contig matrix has not been created. Please run the create_contig_matrix() method first.")
            return None
        if state not in [0,1,2,3]:
            print("state must be 0, 1, 2, or 3")
            return None
        intervals = []
        for idxChr in range(len(self.chrNames)):
            if chromosomeSubset is not None and idxChr not in chromosomeSubset:
                continue
            for idxInd in range(len(self.indNames)):
                if individualSubset is not None and idxInd not in individualSubset:
                    continue
                thisContig = self.contigMatrix[idxChr, idxInd]
                chrIntervals =thisContig.get_my_intervals_of_state(state)
                intervals.extend(chrIntervals)
        return intervals

    def intervals_to_bed(self,outputDir):
        """
        Export intervals of each state to BED files for each chromosome and state.

        Args:
            outputDir (str): Directory to save the BED files.
        """
        if self.contigMatrix is None:
            print("contig matrix has not been created. Please run the create_contig_matrix() method first.")
            return None
        ct.export_contigs_to_ind_bed_files(self,outputDir)
        print("BED files created in directory: ",outputDir)
        return None


# save diemtype function needs to be updated.  The dictionary of variables is saved exactly as below
# however, now the contigMatrix needs to be 'packed' as well in order to be saved correctly.

def save_DiemType(diemTypeObj, pcklPath):
    '''
    Save the dictionary of variables for a DiemType object to a pickle file.
    File extensions something like my_name.diemtype.dict.pkl

    Args:
        diemTypeObj (DiemType): The DiemType object to be saved.
        pcklPath (str): Path to save the pickle file.
    '''
    # the contigMatrix needs to be packaged
    # however, we do not want to overwrite the original contigMatrix
    # so we pack the dictionary version of the contigMatrix
    # and then save that

    d = diemTypeObj.__dict__.copy()
    if d['contigMatrix'] is not None:
        d['contigMatrix'] = ct.pack_contig_matrix(d['contigMatrix'])

    with open(pcklPath, 'wb') as f:
        pickle.dump(d, f,protocol=pickle.HIGHEST_PROTOCOL)


# as in saveDiemType, the contigMatrix needs to be unpacked after loading
def load_DiemType(pcklPath):
    '''
    Load a dict of objects that make a DiemType, then construct that DiemType from the loaded data.
    This helps ensure that, so long as variables are not renamed or removed, DiemType objects can be saved and loaded even if the class definition changes.
    attributes can be added, and we could later add code to handle removed attributes if needed.


    Args:
        pcklPath (str): Path to the pickle file.

    Returns:
        DiemType: The loaded DiemType object.
    '''
    blankArgs = [None for x in range(6)]
    d = DiemType(*blankArgs)

    with open(pcklPath, 'rb') as f:
        d.__dict__.update(pickle.load(f))

    if d.contigMatrix is not None:
        d.contigMatrix = ct.unpack_contig_matrix(d.contigMatrix)

    return d


import numba
import numpy as np
#import timeit
#import time
import math
from matplotlib import pyplot as plt
#import pandas as pd
from numba import prange
import multiprocessing as mp



# this function returns the reach, i.e. distance away that we wil look at given the scale
@numba.njit
def TruncatedLaplaceReach(scale): 
    return scale * 2.995732273553991

# this function returns the weight using the truncated laplace (unnormalized, as in stuart's notebook)
@numba.njit
def TruncatedLaplace(x,scale):
    
    if (0<= x) and ( x < scale*2.995732273553991):
        res = 0.5263157894736842 * math.exp( - x / scale)
    elif (x<0) and (-scale * 2.995732273553991 < x):
         res = 0.5263157894736842 * math.exp( x / scale)
    else: res = 0

    return res
    
# this function returns the weights of each site using TruncatedLaplace defined above
@numba.njit
def get_laplace_weights(posArr,focus,scale):
    if len(posArr) < 1: raise Exception("the number of positions must be >=1")
    posArr = posArr - posArr[focus]
    
    #get the weights of each site
    weights = np.zeros(len(posArr))-1.0
    for idx in range(len(weights)):
        weights[idx] = TruncatedLaplace(posArr[idx],scale)

    return weights




# this function gets the weights of all the markers within the window
# it returns the new state after smothing, determined by the max weight or the nearest marker with the max weight (including itself)
# *** this version WILL convert a 'called' marker to an 'uncalled' marker if the surrounding markers are 'uncalled'
@numba.njit
def get_laplace_smoothed_value(posArr,valArr,focus,weights): #this version will convert a called marker to an uncalled marker
    #print('val arr = ', valArr)                         
    res = -1
    if len(valArr)<1: raise Exception(f"value array must be non-empty", f"focus = {focus}", "pos array = ", posArr , "val array = ", valArr, " weights = ", weights) 
    if len(valArr) == 1: #if there are no other informative markers nearby, return the current value
        return valArr[0]
        
    posArr = posArr - posArr[focus] #change to distances relative to focal site

    # create an array with the weights of the markers [0, 1, 2]
    zeroWeight = 0+sum(weights[valArr == 0])
    oneWeight = 0+sum(weights[valArr==1])
    twoWeight = 0+sum(weights[valArr == 2])
    threeWeight = 0+sum(weights[valArr == 3])
    weightArr = np.array([zeroWeight,oneWeight,twoWeight,threeWeight]) 

    # get the states with the higest weight
    mVal = np.max(weightArr)
    mIndex = np.array([0,1,2,3])[weightArr == mVal]

    if len(mIndex == 1): #there is an obvious max
        res = mIndex[0]
    else: # it could go more than one way
        if np.any(mIndex == valArr[focus]): #if the current value could be the max, do not reassign
            res = valArr[focus]
        else: #assign the state of the closest marker with the maximum weight, and if that fails (it shouldn't), do not reassign
            xv = valArr[focus]
            xd = max(posArr)+1
            for m in mIndex:
                minDist = min(posArr[valArr == m])
                if minDist < xd:
                    xd = minDist
                    xv = m
            res = xv
    
    if res == -1: raise Exception("something went wrong with get_laplace_smoothed_value")
        
    return res


@numba.njit
def get_interval_indices(posArr,scale):
    maxDist = TruncatedLaplaceReach(scale)
    #print("scale is ",scale," and max distance is ",maxDist) 
    
    left = 0
    focus = 0
    right = 0
    
    nSites = len(posArr)
    res = np.zeros((nSites,2),dtype=np.int64)
    

    for focusIdx,focusPos in enumerate(posArr):

        while left<=focusIdx and (focusPos - posArr[left])>maxDist: left +=1
        while right< len(posArr)-1 and (posArr[right+1] - focusPos)<maxDist: right += 1

        res[focusIdx] = np.array([left,right],dtype=np.int64)
    return res



@numba.njit
def laplace_smooth_multiple_haplotypes(posArr,valArr,scale): #this version does change the state of uncalled markers
    #posArr is still a 1D array
    #valArr is now multi-dimensional.  The rows correspond to haplotypes and the columns to positions

    maxDist = TruncatedLaplaceReach(scale/2)

    # print("scale is ",scale," and max distance is ",maxDist)
    
    left = 0
    focus = 0
    right = 0

    nHaplotypes = len(valArr)
    nSites = len(posArr)



    # if nHaplotypes >= nSites:
    #     print("there are more haplotypes than sites in this dataset. check that you have passed the  position array and value array in the correct order")
    #     print("continuing evaluation as if everything were normal")
    #if nHaplotypes == nSites: raise Exception("number of haploptypes and sites are the same. Make sure valArr is a 2D array with haplotypes as rows and marker states as columns")
    res = np.zeros((nHaplotypes,nSites),dtype=np.int8)-1
    
    for focusIdx,focusPos in enumerate(posArr):
        # print(focusIdx)
        # if focusIdx>10:
        #     break

        
        while left<=focusIdx and (focusPos - posArr[left])>maxDist: left +=1
        while right< len(posArr)-1 and (posArr[right+1] - focusPos)<maxDist: right += 1
        weights = get_laplace_weights(posArr[left:right+1],focusIdx - left,scale)
        #print(weights)
        #print(left,focusIdx,right)
        # if len(valArr[0][left:right+1]) == 0:
        #     print('focus Idx = ', focusIdx, 'left = ', left, 'right = ', right)
        #     print(valArr[:,left:right+1])
        #     break
        for haplotypeIndex in range(nHaplotypes):
            
            #print(valArr[haplotypeIndex][left:right+1])
            # uncomment this  and comment out the other lines to make it keep all Us as Us
            # if valArr[haplotypeIndex][focusIdx] == 3: #if this haplotype specifically has an uncalled marker, don't change anything
            #     v = 3
            # else:
            #     v = get_laplace_smoothed_value(posArr[left:right+1],valArr[haplotypeIndex][left:right+1],focusIdx - left,weights)
            #print('val arr = ', valArr[haplotypeIndex][left:right+1])
            v = get_laplace_smoothed_value(posArr[left:right+1],valArr[haplotypeIndex][left:right+1],focusIdx - left,weights)
            #print(f'v = {v}')
            res[haplotypeIndex][focusIdx] = v
            
    return res        


@numba.njit(parallel=True)
def laplace_smooth_multiple_haplotypes_parallel(posArr, valArr, scale):
    """
    Parallel version of laplace_smooth_multiple_haplotypes using numba prange.
    Parallelizes across genomic sites for better performance.
    """
    maxDist = TruncatedLaplaceReach(scale/2)
    
    nHaplotypes = len(valArr)
    nSites = len(posArr)
    
    # if nHaplotypes >= nSites:
    #     print("there are more haplotypes than sites in this dataset. check that you have passed the position array and value array in the correct order")
    #     print("continuing evaluation as if everything were normal")
    
    res = np.zeros((nHaplotypes, nSites), dtype=np.int8) - 1
    
    # Pre-compute intervals for all sites (this helps with parallelization)
    # IMPORTANT: Use scale/2 to match the sequential version
    intervals = get_interval_indices(posArr, scale/2)
    
    # Parallelize across sites using prange
    for focusIdx in prange(nSites):
        left = intervals[focusIdx][0]
        right = intervals[focusIdx][1]
        
        # Get weights for this site's window
        weights = get_laplace_weights(posArr[left:right+1], focusIdx - left, scale)
        
        # Process all haplotypes for this site
        for haplotypeIndex in range(nHaplotypes):
            v = get_laplace_smoothed_value(
                posArr[left:right+1], 
                valArr[haplotypeIndex][left:right+1], 
                focusIdx - left, 
                weights
            )
            res[haplotypeIndex][focusIdx] = v
    
    return res



#!/usr/bin/env python3

"""vcf2diembeds.py

Usage:
  vcf2diembeds.py <vcf_file_path> [--outdir DIR]

Examples:
  vcf2diembeds.py path/to/file.vcf
  vcf2diembeds.py path/to/file.vcf --outdir results/

  
Features:
  vcf files of any size (eg 26Gb filesize, 474 genomes, 3.6 million sites, 37 mins (see increment...now 10 mins?)
  lossless encoding of vcf variants of up to 4 states (diptypes with further states are U-encoded)
  outputs 3 bed-like files (TSV with #-commented header)
       1 file.vcf.diem_input.bed      variant input for diem
       2 file.vcf.diem_exclude.bed    vcf sites excluded for diem
       3 file.vcf.diem_meta.bed       default metadata input for diem

  The new release of diem.py should run on {1,3}
  Ideally 3 will first be updated by the user before running diem
  (using 'update_diem_metadata' in prep)

Development: (Stuart J.E. Baird)
    2. increment: extended from '0123456789' encoding to DIEM alphabet (55 symbols; 0–9,a–z,A–S)
       This means 'main' has some stuff before the bit that resembles the original chatGPT5 vanilla VCF interpreter
    1. increment: Optimised file I/O (4x speedup).
    
    0. Inspired by vcf2diem.py by Sam Ebdon this is, however, a complete re-write.
    I asked chatGPT5 to write me a vanilla vcf interpreter (here, ± == main)
    I then added 5 functions (immediately below) and tidied.
    The result is much shorter and simpler than vcf2diem.py, with fewer dependencies.
    This is to be expected, as it does a different job: here there is minimal filtering.
    In the emBEDed diem workflow filtering experiments should be downstream of the [slow: big files] VCF interpreter.
"""

import sys, os
from timeit import default_timer as timer
from collections import Counter
from datetime import timedelta
import io, gzip
import operator
import itertools  # used by diem_most_common_alleles (downstream tool)
import re
import argparse


def open_text(path):
    if path.endswith(".gz"):
        # For reading, compresslevel is ignored; use TextIOWrapper for decoding
        return io.TextIOWrapper(gzip.open(path, "rb"), encoding="utf-8", newline="")
    return open(path, "r", encoding="utf-8", buffering=1 << 20, newline="")

def open_out(path):
    return open(path, "w", encoding="utf-8", buffering=1 << 20, newline="")

def convert_seconds(n):
    return str(timedelta(seconds=n))

#---------------- DIEM alphabet (55 symbols) start -------------------------#
diemALPHABET = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRS'
diemUencodableChar = '_'
diemMaxVariants = 10
diemMaxChars = diemMaxVariants * (diemMaxVariants + 1) // 2  # 55

def diemGentobase62(n):
    if 0 <= n < diemMaxChars:
        return diemALPHABET[n]
    else:
        return '!'  # Error case

# used once to create a lookup table
def base62ordConversion(n):
    if 48 <= n <= 57:
        return n - 48  # '0'..'9' -> 0..9
    elif 65 <= n <= 90:
        return n - 29  # 'A'..'Z' -> 36..61 (we only use up to 'S')
    elif 97 <= n <= 122:
        return n - 87  # 'a'..'z' -> 10..35
    else:
        return -1  # Error case

# a lookup tuple: ord(char) -> diemGen index (or -1)
base62todiemGenLookUp = tuple(base62ordConversion(n) for n in range(123))
def base62todiemGen(s):
    if len(s) == 1:
        n = ord(s)
        if 0 <= n <= 122:
            return base62todiemGenLookUp[n]
    return -1  # Error case
   
# two small lookup tuples ensure compatibility with 0123456789 encoding without runtime cost
ABSijLookupTable = (0, 0, 5, 5, 5, 9, 11, 13, 15, 17)
MAXijLookupTable = (0, 0, 0, 0, 3, 6, 10, 15, 21, 28)
def diem_encode_allele_ranks(i, j):
    maxij = max(i, j)
    if maxij >= diemMaxVariants:
        return diemUencodableChar
    absij = abs(i - j)
    return diemGentobase62(i + j + ABSijLookupTable[absij] + MAXijLookupTable[maxij])  # see lossless encoding slideset

# construct decode lookup
diem_decode_allele_ranks_lookup = [[base62todiemGen(diem_encode_allele_ranks(i, j)), [i, j]]
                                   for j in range(diemMaxVariants) for i in range(diemMaxVariants) if i <= j]
diem_decode_allele_ranks_lookup.sort(key=lambda x: x[0])
diem_decode_allele_ranks_lookup[:] = [row[1] for row in diem_decode_allele_ranks_lookup]

def diem_decode_allele_ranks(s):
    diemgen = base62todiemGen(s)
    if 0 <= diemgen < diemMaxChars:
        return diem_decode_allele_ranks_lookup[diemgen]
    else:
        return []

def diem_most_common_alleles(site_string):  # tool for downstream
    gtypes = [diem_decode_allele_ranks(c) for c in site_string]
    c = Counter(list(itertools.chain(*gtypes)))
    return c.most_common()
#---------------- DIEM alphabet end ---------------------------#

exclusion_NOT = 'E0'   # NOT excluded
exclusion_invar = 'E1' # invariant is simply excluded    

def diem_encode_HOMsingleton_exclusion(diem_gtypes: str) -> str:
    """O(1) string ops: counts & first indices; no per-char lists."""
    lz = diem_gtypes.count('0')
    lt = diem_gtypes.count('2')
    if lz > 1 and lt > 1:
        return exclusion_NOT
    if lz == 0 and lt == 0:
        return 'E2'  # ZERO HOMs
    if lz == 1 and lt == 1:
        zi = diem_gtypes.find('0')
        ti = diem_gtypes.find('2')
        return f'E5_{zi}_{ti}'  # TWO SINGLETON HOMS
    if lz == 0 or lt == 0:
        return 'E3'  # ONLY ONE HOM
    # One singleton HOM case
    idx = diem_gtypes.find('0') if lz == 1 else diem_gtypes.find('2')
    return f'E4_{idx}'

def fastGTs(sample_fields, seqLabelsString, gt_index):
    """Hot path: derive DIEM string & metadata directly from sample_fields."""
    hide_exclusions = False  # internal flag
    diem_enc_a_ranks = diem_encode_allele_ranks
    diem_dec_a_ranks = diem_decode_allele_ranks
    
    seqLabels = seqLabelsString.split(',')
    GTlabelcount = [0] * diemMaxVariants

    def iter_GTs():
        """Yield GT strings with minimal splitting/allocation."""
        if gt_index == 0:
            for s in sample_fields:
                yield s.split(':', 1)[0] if s else './.'
        elif gt_index > 0:
            for s in sample_fields:
                # split up to gt_index to avoid trailing allocations
                parts = s.split(':', gt_index)
                yield parts[gt_index] if len(parts) > gt_index else './.'
        else:
            for _ in sample_fields:
                yield './.'

    def GT(s):
        # normalize phased to unphased once
        if '|' in s:
            s = s.replace('|', '/')
        match s:
            case '0/0':
                GTlabelcount[0] += 2; return '0'
            case '0/1':
                GTlabelcount[0] += 1; GTlabelcount[1] += 1; return '1'
            case '1/1':
                GTlabelcount[1] += 2; return '2'
            case _:
                a, _, b = s.partition('/')
                if a == '.' or b == '.' or not b:
                    return diemUencodableChar
                i = int(a); j = int(b)
                if max(i, j) >= diemMaxVariants:
                    return diemUencodableChar
                GTlabelcount[i] += 1; GTlabelcount[j] += 1
                return diem_enc_a_ranks(i, j)

    prelim_states = [GT(g) for g in iter_GTs()]

    # ---- Early return: invariant / mono-allelic -> skip sorting & translation
    nonzero = diemMaxVariants - GTlabelcount.count(0)
    if nonzero < 2:
        if nonzero == 0:
            usedSeqs = [diemUencodableChar]  # no observed allele labels
        else:
            # find the only allele index with a nonzero count
            k = next(i for i, cnt in enumerate(GTlabelcount) if cnt)
            usedSeqs = [seqLabels[k]]
        SNV = 0  # not a variant
        return nonzero, ','.join(usedSeqs), SNV, exclusion_invar, 'S'
    # ----
    
    simpleOrder = list(range(diemMaxVariants))
    # top-N variant labels by count
    sorted_count_pos = sorted(
        zip(GTlabelcount, simpleOrder), key=lambda x: x[0], reverse=True
    )[:nonzero]
    sorted_count_pos = [pos for _, pos in sorted_count_pos]

    ztExchangeableOrder = [1, 0, 2, 3, 4, 5, 6, 7, 8, 9][:nonzero]

    if (sorted_count_pos == simpleOrder[:nonzero]
        or sorted_count_pos == ztExchangeableOrder):  # NO reordering necessary
        usedSeqs = seqLabels[:nonzero]
        SNV = int(max(map(len, usedSeqs)) == 1)
        diemString = 'S' + ''.join(prelim_states)
        exclusion_crit = diem_encode_HOMsingleton_exclusion(diemString)
    else:  # reordering possibly necessary (not checking tie possibilities), rare
        getter = operator.itemgetter(*sorted_count_pos)
        picked = getter(seqLabels)
        usedSeqs = list(picked) if isinstance(picked, tuple) else [picked]
        SNV = int(max(map(len, usedSeqs)) == 1)

        if nonzero == 1:
            diemString = 'S'
            exclusion_crit = exclusion_invar
        else:
            new_ranks = getter(simpleOrder)
            if not isinstance(new_ranks, tuple):
                new_ranks = (new_ranks,)
            revhash = [-1] * (max(new_ranks) + 1)
            for i, pos in enumerate(new_ranks):
                revhash[pos] = i

            oldstates = set(prelim_states)
            oldrankpairs = [[s, diem_dec_a_ranks(s)] for s in oldstates if s != diemUencodableChar]
            replacements = [(s, diem_enc_a_ranks(revhash[rp[0]], revhash[rp[1]]))
                            for s, rp in oldrankpairs]

            prelimdiemString = 'S' + ''.join(prelim_states)
            chr_translation = str.maketrans(dict(replacements))
            diemString = prelimdiemString.translate(chr_translation)
            exclusion_crit = diem_encode_HOMsingleton_exclusion(diemString)

    if exclusion_crit != exclusion_NOT and hide_exclusions:  # hide excluded sites' strings
        diemString = 'S'

    return nonzero, ','.join(usedSeqs), SNV, exclusion_crit, diemString


def main():
    parser = argparse.ArgumentParser(
        prog=os.path.basename(sys.argv[0]),
        description="Convert a VCF(.gz) into DIEM bed-like outputs."
    )
    parser.add_argument("vcf_file_path", help="Path to input VCF (.vcf or .vcf.gz)")
    parser.add_argument(
        "--outdir",
        "-o",
        default=None,
        help="Optional output directory. Defaults to the input file's directory."
    )
    args = parser.parse_args()

    file_path = args.vcf_file_path
    start_time = timer()

    # Decide output directory
    if args.outdir is None:
        outdir = os.path.dirname(os.path.abspath(file_path)) or "."
    else:
        outdir = os.path.abspath(args.outdir)
        os.makedirs(outdir, exist_ok=True)

    # Base name for outputs (keep full name incl .vcf/.vcf.gz, matching your current naming)
    base = os.path.basename(file_path)

    variant_output_path = os.path.join(outdir, base + ".diem_input.bed")
    exclude_output_path = os.path.join(outdir, base + ".diem_exclude.bed")
    meta_output_path    = os.path.join(outdir, base + ".diem_meta.bed")


    samples = []
    chromosome_tally = Counter()
    minpos = {}   # chrom -> min start
    maxpos = {}   # chrom -> max end
    contig_lengths = {}  # chrom -> reference length (1-based length; end-exclusive in 0-based)
    nvars = necls = 0

    with open_text(file_path) as vcf, \
         open_out(variant_output_path) as vout, \
         open_out(exclude_output_path) as xout:

        invariant_header = '#Chrom\tStart\tEnd\tQual\tRef\tSeqAlleles\tSNV\tnVNTs\tExclusionCriterion\t'
        # headers
        vout.write(invariant_header + '|'.join([]) + '\n')  # will patch later after samples known
        xout.write(invariant_header + 'Dstring\n')

        gt_index = None
        last_fmt = None

        for line in vcf:
            if line.startswith('##contig='):
                m = re.search(r'ID=([^,>]+).*?length=(\d+)', line)
                if m:
                    cid, clen = m.group(1), int(m.group(2))
                    contig_lengths[cid] = clen
                continue
            if line.startswith('##'):
                continue
            if line.startswith('#CHROM'):
                header_fields = line.rstrip('\n').split('\t')
                samples = header_fields[9:]
                # rewrite variant header now that we know samples
                vout.seek(0)
                vout.write(invariant_header + '|'.join(samples) + '\n')
                vout.seek(0, io.SEEK_END)
                continue

            fields = line.rstrip('\n').split('\t')
            chrom = fields[0]
            pos   = fields[1]
            start = str(int(pos) - 1)   # VCF 1-based -> BED 0-based
            end   = pos                 # single-site interval description
            ref   = fields[3]
            alt   = fields[4]
            qual  = fields[5]
            fmt   = fields[8]
            sample_fields = fields[9:]

            # compute GT column index once per site-format (handle changing FORMAT)
            if gt_index is None or fmt != last_fmt:
                fmt_cols = fmt.split(':')
                gt_index = fmt_cols.index('GT') if 'GT' in fmt_cols else -1
                last_fmt = fmt

            # Interpret GTs fast (no intermediate genotypes list)
            nVNTs, ordered_SAs, SNV, exclusion_criterion, dg = fastGTs(sample_fields, ref + ',' + alt, gt_index)
            
            if exclusion_criterion == exclusion_NOT:
                vout.write(f"{chrom}\t{start}\t{end}\t{qual}\t{ref}\t{ordered_SAs}\t{SNV}\t{nVNTs}\t{exclusion_criterion}\t{dg}\n")
                nvars += 1
                chromosome_tally[chrom] += 1
                if chrom not in minpos or int(start) < int(minpos[chrom]): minpos[chrom] = start
                if chrom not in maxpos or int(end)   > int(maxpos[chrom]): maxpos[chrom] = end
            else:
                xout.write(f"{chrom}\t{start}\t{end}\t{qual}\t{ref}\t{ordered_SAs}\t{SNV}\t{nVNTs}\t{exclusion_criterion}\t{dg}\n")
                necls += 1

    # Write META (once, buffered) -- all 0-based
    default_ploidy = ['2'] * len(samples)
    default_rec_rate = 1.0

    with open_out(meta_output_path) as mout:
        mout.write(
            '#Chrom\tRefStart0\tRefEnd0\tFirstVariantStart0\tLastVariantPos0\t'
            'n(diem_inputs)\trelativeRecRates\t' + '\t'.join(samples) + '\n'
        )
        for chrom, count in chromosome_tally.items():
            ref_start0 = '0'
            # BED-style end-exclusive: if length is L, RefEnd0 = L
            ref_end0 = str(contig_lengths.get(chrom, ''))  # empty if unknown
            first_var_start0 = minpos[chrom]               # already 0-based in your code
            last_variant_pos0 = str(int(maxpos[chrom]) - 1)  # convert 1-based VCF to 0-based
            mout.write(
                f"{chrom}\t{ref_start0}\t{ref_end0}\t{first_var_start0}\t{last_variant_pos0}\t"
                f"{count}\t{default_rec_rate}\t" + '\t'.join(default_ploidy) + '\n'
            )
            
    total = nvars + necls
    print(f"{total:,d} vcf sites for {len(samples)} genomes, split to {nvars:,d} diem variants with {necls:,d} excluded. ({round(100 * necls / max(total,1))}% reduction).")
    print(f"diem_meta.bed data saved to {meta_output_path}")
    print(f"diem_input.bed data saved to {variant_output_path}")
    print(f"diem_exclude.bed data saved to {exclude_output_path}")
    print(f"Time taken (hours:mins:secs): {convert_seconds(timer() - start_time)}")

if __name__ == "__main__":
    main()



# --- diem2fasta: clean notebook-friendly rewrite ---

"""
diem2fasta

Convert vcf2diem outputs (diem_input.bed + optional diem_exclude.bed) into:
- FASTA reconstructions per reference scaffold (or per requested regions),
- per-sample heterozygosity summary,
- pairwise distance matrices computed from DIEM site-state strings (Dstrings).

Core assumptions / semantics:

Dstrings
- Each row describes one site; the Dstring begins with 'S' then one character per sample
  in the META/VCF sample order.
- The DIEM alphabet encodes diplotypes formed from pairs of the ten most common alleles at a site.
- Unencodable state: any site-state character that is '_' or 'U', or any character not
  decodable by the DIEM decode table. This includes:
    (i) diplotypes involving alleles outside the ten most common, and
    (ii) missing/no-data.
- Unencodable calls do not contribute to distance coverage.

Reference / coordinates
- Reference scaffold names and lengths are taken from the provided reference FASTA.
- META is treated as auxiliary metadata (e.g., sample order/ploidy) and is not used to
  define what regions are converted to FASTA.
- If regions_bed_path is provided, diem2fasta writes ONLY the region FASTA(s) described
  there (labels preserved); otherwise it writes full-length per-scaffold FASTAs.

Distances
- Unphased Hamming with partial credit: distance is computed per site from decoded diplotypes,
  accounting for unknown phase; unencodable calls are skipped (coverage counts computed sites).
- Compatibility distance: distance 0 means the pair could be consistent with the same underlying
  sequence given missing/unencodable calls; conflicts are scored from decoded diplotypes.
"""


from __future__ import annotations
import os, sys, math
from typing import Dict, List, Tuple, Optional
import numpy as np
import pandas as pd
import pysam
from multiprocessing import Pool, cpu_count

# Treat BOTH '_' and 'U' as standard “unencodable”, plus anything unknown.

def build_diem_decoder():
    """
    Build DIEM char -> (i,j) map for i<=j for the 10-allele model.
    Returns:
      decode: dict[str, tuple[int,int] | None]
      is_unencodable_char: callable(ch)->bool, treating '_'/'U'/unknown as unencodable
    """
    diemALPHABET = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRS'
    diemMaxVariants = 10
    diemMaxChars = diemMaxVariants * (diemMaxVariants + 1) // 2
    ABSij = (0, 0, 5, 5, 5, 9, 11, 13, 15, 17)
    MAXij = (0, 0, 0, 0, 3, 6, 10, 15, 21, 28)

    def enc(i, j):
        if max(i, j) >= diemMaxVariants:
            return None
        n = i + j + ABSij[abs(i - j)] + MAXij[max(i, j)]
        if 0 <= n < diemMaxChars:
            return diemALPHABET[n]
        return None

    decode = {}
    for j in range(diemMaxVariants):
        for i in range(j + 1):
            ch = enc(i, j)
            if ch is not None:
                decode[ch] = (i, j)

    # Standard “unencodable” sentinels
    decode["_"] = None
    decode["U"] = None

    UNENC = {"_", "U"}

    def is_unencodable_char(ch: str) -> bool:
        # Anything unknown OR sentinel is unencodable.
        # (Unknown chars: decode.get -> None)
        return (ch in UNENC) or (decode.get(ch, None) is None)

    return decode, is_unencodable_char

    
# -------------------- IUPAC helpers --------------------

IUPAC_HET = {
    frozenset(("A","G")): "R",
    frozenset(("C","T")): "Y",
    frozenset(("G","C")): "S",
    frozenset(("A","T")): "W",
    frozenset(("G","T")): "K",
    frozenset(("A","C")): "M",
}
def iupac_from_pair(a: str, b: str) -> str:
    """Return IUPAC for unordered (a,b) DNA; identical -> a; unknown -> 'N'."""
    if a == b:
        return a
    return IUPAC_HET.get(frozenset((a,b)), "N")

# -------------------- IO: meta / bed --------------------

def load_meta(meta_path: str,
              ref_fasta_path: str,
              variants_path: str | None = None):
    """
    META reader for diem2fasta:

    • If META has RefStart0/RefEnd0 -> use them for lengths (new format).
    • Otherwise (old META with Start_diem_input/End_diem_input) ignore those for coords
      and infer 0..len(FASTA[chrom]).
    • META is still used for scaffold list/order (#Chrom) and sample columns/order if present.
    • If META lacks samples, infer sample order from the last header column of the variants BED
      (pipe-joined names).
    """
    import pandas as pd, numpy as np, pysam

    # --- FASTA is the fallback (and the truth for old META)
    fa = pysam.FastaFile(ref_fasta_path)
    fa_order = list(fa.references)
    fa_len   = {r: fa.get_reference_length(r) for r in fa_order}

    # --- Read META
    df = pd.read_csv(meta_path, sep="\t", dtype="string")

    # #Chrom (case tolerant)
    chrom_col = None
    for c in df.columns:
        if c.lower() in ("#chrom", "chrom"):
            chrom_col = c; break
    if chrom_col is not None:
        df = df.rename(columns={chrom_col: "#Chrom"})
        chr_names = df["#Chrom"].astype(str).to_numpy()
    else:
        # fallback if META doesn't list chroms
        if variants_path:
            # minimal scan to get a chrom column from variants
            nskip = 0
            with open(variants_path, "r", encoding="utf-8") as fh:
                for line in fh:
                    if line.startswith("##"): nskip += 1; continue
                    if line.startswith("#"):  nskip += 1
                    break
            dv = pd.read_csv(variants_path, sep="\t", dtype="string", engine="python",
                             nrows=200000, skiprows=max(nskip-1, 0))
            chrom_alt = None
            for c in dv.columns:
                if c.lower() in ("#chrom", "chrom"):
                    chrom_alt = c; break
            if chrom_alt is not None:
                chr_names = np.array(sorted(dv[chrom_alt].dropna().astype(str).unique()), dtype=object)
            else:
                chr_names = np.array(fa_order, dtype=object)
        else:
            chr_names = np.array(fa_order, dtype=object)

    # --- Decide how to compute lengths
    # Prefer explicit RefStart0/RefEnd0 if present; otherwise infer from FASTA.
    refstart = next((c for c in df.columns if c.lower() == "refstart0"), None)
    refend   = next((c for c in df.columns if c.lower() == "refend0"),   None)

    if refstart and refend:
        # Newer META: trust these values
        s = pd.to_numeric(df[refstart], errors="coerce").fillna(0).astype("int64").to_numpy()
        e = pd.to_numeric(df[refend],   errors="coerce").to_numpy()
        # align lengths to chr_names order
        len_map = {}
        for i in range(len(df)):
            c = str(df.loc[i, "#Chrom"]) if "#Chrom" in df.columns else None
            if c is not None and pd.notna(e[i]):
                len_map[c] = int(e[i]) - int(s[i])
        chr_lengths = np.array([int(len_map.get(str(c), fa_len.get(str(c), 0)))
                                for c in chr_names], dtype="int64")
    else:
        # Old META: ignore Start_diem_input/End_diem_input; use reference lengths
        chr_lengths = np.array([int(fa_len.get(str(c), 0)) for c in chr_names], dtype="int64")

    # --- Samples & ploidy
    # Heuristic: everything to the right of the last known fixed meta field is samples.
    fixed_candidates = [
        "n(diem_inputs)", "lastvariantpos0", "firstvariantstart0",
        "end_diem_input", "start_diem_input", "end"
    ]
    low = {c.lower(): c for c in df.columns}
    last_fixed_idx = -1
    for key in fixed_candidates:
        if key in low:
            last_fixed_idx = max(last_fixed_idx, list(df.columns).index(low[key]))

    if last_fixed_idx >= 0 and last_fixed_idx + 1 < len(df.columns):
        sample_names = df.columns[last_fixed_idx+1:].to_numpy()
        # per-chrom ploidy, default to 2 where missing
        ploidy = df.iloc[:, last_fixed_idx+1:].astype("Int64").fillna(2).to_numpy()
        # align ploidy to chr_names if #Chrom present
        if "#Chrom" in df.columns and len(ploidy) == len(df):
            row_map = {str(df.loc[i, "#Chrom"]): i for i in range(len(df))}
            ploidy = np.vstack([
                (ploidy[row_map[c]] if c in row_map else np.full(len(sample_names), 2, dtype=int))
                for c in map(str, chr_names)
            ])
    else:
        # No sample block in META → infer from variants header (pipe-joined)
        if not variants_path:
            raise ValueError("Cannot infer sample names: META has no sample columns and variants_path not provided.")
        last_header = None
        with open(variants_path, "r", encoding="utf-8") as fh:
            for line in fh:
                if line.startswith("#Chrom") or line.startswith("#CHROM"):
                    last_header = line.rstrip("\n").split("\t")[-1]
                    break
        if last_header is None:
            raise ValueError("Could not find variants header to infer sample names.")
        sample_names = np.array(last_header.split("|"), dtype=object)
        ploidy = np.full((len(chr_names), len(sample_names)), 2, dtype="int64")

    return chr_names, chr_lengths, sample_names, ploidy

def read_regions_bed(path):
    """
    Read BED of regions to export.

    Columns:
        chrom  start  end  [label...]

    The 4th column may contain spaces — everything after column 3 is
    treated as the label verbatim.

    Returns list of tuples:
        (chrom:str, start:int, end:int, label:str)
    """
    regions = []

    with open(path, "r", encoding="utf-8") as fh:
        for ln, line in enumerate(fh, 1):
            line = line.rstrip("\n")

            if not line or line.startswith("#"):
                continue

            # Split only first 3 columns
            parts = line.split(maxsplit=3)

            if len(parts) < 3:
                raise ValueError(f"{path}:{ln} — fewer than 3 columns")

            chrom = parts[0]
            start = int(parts[1])
            end   = int(parts[2])

            # Label = remainder verbatim if present
            if len(parts) == 4:
                label = parts[3]
            else:
                label = f"{chrom}:{start}-{end}"

            regions.append((chrom, start, end, label))

    return regions


def read_bed_as_strings(path: str, skip_preamble: bool=True
                       ) -> pd.DataFrame:
    """
    Read a DIEM BED-like TSV as all strings. The last column is the Dstring
    and is renamed to 'diem_genotype'. Works for both variants and excludes.
    """
    # Count preamble (##...) and header (#Chrom...) lines
    nskip = 0
    with open(path, "r", encoding="utf-8") as fh:
        for line in fh:
            if line.startswith("##"):
                nskip += 1
                continue
            if line.startswith("#"):
                nskip += 1
            break

    df = pd.read_csv(path, sep="\t", dtype="string", engine="python", skiprows=nskip-1)
    # last column contains D-string with very long header (pipe-joined samples)
    lastcol = df.columns[-1]
    df = df.rename(columns={lastcol: "diem_genotype"})
    # normalize key columns
    for c in ("chrom","Chrom","#Chrom"):
        if c in df.columns:
            df = df.rename(columns={c: "chrom"})
            break
    if "exclusion_criterion" not in df.columns:
        for c in df.columns:
            if c.lower() == "exclusioncriterion":
                df = df.rename(columns={c:"exclusion_criterion"})
                break
    return df

def union_variants(df_v: pd.DataFrame, df_x: Optional[pd.DataFrame]) -> pd.DataFrame:
    """Keep all rows; excludes are useful except pure invariants E1."""
    if df_x is None:
        return df_v.copy()
    df_all = pd.concat([df_v, df_x], ignore_index=True)
    # normalize exclusion codes and drop E1
    if "exclusion_criterion" in df_all.columns:
        df_all["exclusion_criterion"] = df_all["exclusion_criterion"].astype("string").str.strip()
        df_all = df_all[df_all["exclusion_criterion"] != "E1"].copy()
    return df_all


def _normalize_bed_columns_inplace(df):
    """Map legacy BED headers to canonical names used downstream."""
    if df is None or df.empty:
        return df
    low = {c.lower(): c for c in df.columns}

    def have(*opts):
        for o in opts:
            if o in low:
                return low[o]
        return None

    ren = {}
    if have('#chrom','chrom'): ren[have('#chrom','chrom')] = 'chrom'
    if have('start','start_diem_input'): ren[have('start','start_diem_input')] = 'start'
    if have('end','end_diem_input'):     ren[have('end','end_diem_input')]     = 'end'
    if have('qual'): ren[have('qual')] = 'qual'
    if have('ref'):  ren[have('ref')]  = 'ref'
    if have('seqalleles'): ren[have('seqalleles')] = 'SeqAlleles'
    if have('snv'):        ren[have('snv')]        = 'SNV'
    if have('nvnts','n(vnts)','n_vnts','nvariants'): ren[have('nvnts','n(vnts)','n_vnts','nvariants')] = 'nVNTs'
    if have('exclusion_criterion','exclusioncriterion'):
        ren[have('exclusion_criterion','exclusioncriterion')] = 'exclusion_criterion'
    if have('diem_genotype','dstring','dstrings'):
        ren[have('diem_genotype','dstring','dstrings')] = 'diem_genotype'

    df.rename(columns=ren, inplace=True)
    return df

    
# -------------------- Dstring & site helpers --------------------

def assert_dstring_shape(df: pd.DataFrame, n_samples: int, where: str="variants"):
    """Ensure diem_genotype is present and each row starts with 'S' and len==n+1."""
    if "diem_genotype" not in df.columns:
        raise ValueError(f"{where}: missing 'diem_genotype' column (check last column rename).")
    ok = df["diem_genotype"].str.startswith("S") & (df["diem_genotype"].str.len() == n_samples + 1)
    if not ok.any():
        # Give a short hint
        lens = df["diem_genotype"].str.len().value_counts().head(5).to_dict()
        raise ValueError(f"{where}: no valid Dstrings: startswith('S')={df['diem_genotype'].str.startswith('S').mean():.3f}, "
                         f"lengths(top)={lens}, expected={n_samples+1}")

def rows_by_chrom(df: pd.DataFrame, chroms: np.ndarray) -> Dict[str, pd.DataFrame]:
    out = {}
    for c in chroms:
        sub = df[df["chrom"].astype(str) == str(c)].copy()
        out[str(c)] = sub
    return out

# -------------------- Reconstruction --------------------

def OLDreconstruct_scaffold(ref_fa, scaffold_name, scaffold_len, rows, sample_names):
    """
    Build per-sample sequences for one scaffold using DIEM Dstrings and SeqAlleles.
    Policies:
      - Homozygous SNV: substitute base.
      - Heterozygous SNV: IUPAC code.
      - Homozygous indel (len(allele)!=len(ref)): apply exact replacement (length changes).
      - Heterozygous indel: conservative 'B' — DO NOT change length; leave ref segment
        intact and write a single 'N' at start position to flag ambiguity.
      - '_' (unencodable): skip / leave reference as-is at that site.
      - Overlapping edits: keep the first applied; skip later overlaps (counted).
    Returns: list[str] of reconstructed sequences (one per sample, same order as sample_names).
    """
    import numpy as np

    n = len(sample_names)

    # ---- DIEM decode (must match your distance code) -----------------------
    diemALPHABET = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRS'
    diemMaxVariants = 10
    diemMaxChars = diemMaxVariants * (diemMaxVariants + 1) // 2
    ABSij = (0, 0, 5, 5, 5, 9, 11, 13, 15, 17)
    MAXij = (0, 0, 0, 0, 3, 6, 10, 15, 21, 28)

    def enc(i, j):
        if max(i, j) >= diemMaxVariants: return None
        n_ = i + j + ABSij[abs(i - j)] + MAXij[max(i, j)]
        if 0 <= n_ < diemMaxChars:
            return diemALPHABET[n_]
        return None

    decode = {}
    for j in range(diemMaxVariants):
        for i in range(j + 1):
            ch = enc(i, j)
            if ch is not None:
                decode[ch] = (i, j)
    decode['_'] = None  # missing/unencodable

    # IUPAC for heterozygous SNVs
    IUPAC = {
        frozenset(("A","G")): "R",
        frozenset(("C","T")): "Y",
        frozenset(("G","C")): "S",
        frozenset(("A","T")): "W",
        frozenset(("G","T")): "K",
        frozenset(("A","C")): "M",
        frozenset(("A","C","G")): "V",
        frozenset(("A","C","T")): "H",
        frozenset(("A","G","T")): "D",
        frozenset(("C","G","T")): "B",
        frozenset(("A","C","G","T")): "N",
    }

    def iupac_for(a, b):
        a = a.upper(); b = b.upper()
        if a == b: return a
        return IUPAC.get(frozenset((a,b)), "N")

    # Fetch the reference once
    ref_seq = ref_fa.fetch(reference=scaffold_name, start=0, end=int(scaffold_len))
    # Per-sample builders as lists for efficiency
    builders = [list(ref_seq) for _ in range(n)]

    # Track which reference intervals we’ve already replaced (to avoid collisions)
    # We’ll mark applied intervals on a boolean mask; for long scaffolds this is memory-costly,
    # so we instead keep a list of applied intervals and do a tiny overlap check (counts are small).
    applied_intervals = []  # list of (start0, end0) in current coordinate space per sample?
    # NOTE: coordinates shift after indels. We therefore apply edits sample-by-sample.

    # Sort rows by BED start (0-based)
    rows = rows.sort_values("start", kind="mergesort")  # stable

    # Small counters for a one-line report in your notebook logs
    skipped_overlap = 0
    applied_hom_snv = applied_het_snv = 0
    applied_hom_indel = skipped_het_indel = 0

    # For each sample, apply edits in order
    for s_idx in range(n):
        seq = builders[s_idx]
        # We maintain a running offset that maps original BED coordinates to
        # current sequence coordinates as we change length via indels.
        offset = 0
        applied_intervals.clear()

        for _, r in rows.iterrows():
            ds = str(r.get("diem_genotype", ""))
            if not (ds and ds[0] == "S" and len(ds) == n + 1):
                continue  # malformed; ignore

            ch = ds[s_idx + 1]
            pair = decode.get(ch, None)
            if pair is None:
                continue  # '_' or invalid; leave reference

            start0 = int(r["start"]) + offset  # map to current coord space
            ref_allele = str(r.get("ref", "")) or "N"

            # Parse allele strings by DIEM rank (SeqAlleles is ordered by used ranks)
            seq_labels = str(r.get("SeqAlleles", "")).split(",")
            # Guard: if rank out of range, skip
            max_rank = len(seq_labels) - 1
            (i, j) = pair
            if i > max_rank or j > max_rank:
                continue  # cannot realize without labels

            a_i = seq_labels[i]
            a_j = seq_labels[j]

            # Determine homo/het & SNV/indel nature
            is_hom = (i == j)
            len_ref = len(ref_allele)
            len_i   = len(a_i) if a_i else 0
            len_j   = len(a_j) if a_j else 0

            # Compute the reference interval to replace (current coordinates)
            end0 = start0 + len_ref

            # Overlap check against already-applied intervals in this sample
            # (after coordinate shifts). We keep it simple: if the new [start0,end0)
            # intersects any prior [s,e), we skip this row for this sample.
            collided = False
            for (s,e) in applied_intervals:
                if not (end0 <= s or e <= start0):
                    collided = True; break
            if collided:
                skipped_overlap += 1
                continue

            # SNV if both ref and chosen allele length == 1
            def is_snv_len(x): return len(x) == 1

            if is_hom:
                # Homozygous call — realize exactly
                allele = a_i
                if is_snv_len(ref_allele) and is_snv_len(allele):
                    # SNV: simple substitution
                    seq[start0] = allele.upper() if allele else "N"
                    applied_hom_snv += 1
                    # interval length unchanged; record the 1bp site
                    applied_intervals.append((start0, start0+1))
                else:
                    # INDEL or multi-nt: replace ref segment with allele string
                    # 1) delete ref span
                    del seq[start0:end0]
                    # 2) insert allele
                    ins = list(allele.upper()) if allele else list("N")
                    for k, ch2 in enumerate(ins):
                        seq.insert(start0 + k, ch2)
                    # Update offset for downstream edits
                    delta = len(allele) - len_ref
                    offset += delta
                    applied_hom_indel += 1
                    # Record the *post-edit* occupied interval (use allele length)
                    applied_intervals.append((start0, start0 + len(allele)))
            else:
                # Heterozygous call
                if is_snv_len(ref_allele) and is_snv_len(a_i) and is_snv_len(a_j):
                    # Heterozygous SNV: write IUPAC at single base
                    seq[start0] = iupac_for(a_i, a_j)
                    applied_het_snv += 1
                    applied_intervals.append((start0, start0+1))
                else:
                    # Heterozygous involving indel → conservative 'B':
                    # do not change length; mark ambiguity at start with 'N'
                    seq[start0] = "N"
                    skipped_het_indel += 1
                    applied_intervals.append((start0, start0+1))
        # end for each row
        builders[s_idx] = "".join(seq)
    # end for each sample

    # Tiny summary to help you spot issues in the notebook
    print(
        f"[recon {scaffold_name}] hom_snv={applied_hom_snv}, het_snv={applied_het_snv}, "
        f"hom_indel={applied_hom_indel}, skipped_overlap={skipped_overlap}, "
        f"het_indel_conservative={skipped_het_indel}"
    )

    return builders

def reconstruct_scaffold(ref_fa, scaffold_name, scaffold_len, rows, sample_names):
    """
    Build per-sample sequences for one scaffold using DIEM Dstrings and SeqAlleles.

    Policies:
      - HOM SNV: substitute base.
      - HET SNV: IUPAC code.
      - HOM indel/multibase: exact replacement (length changes).
      - HET indel/multibase: conservative 'N' at start; do NOT change length.
      - Unencodable (including '_' and 'U' and any unknown site-state char): skip (leave ref).
      - Overlapping edits: keep first; skip later overlaps.
    """
    n = len(sample_names)

    decode, is_unencodable_char = build_diem_decoder()

    IUPAC = {
        frozenset(("A","G")): "R",
        frozenset(("C","T")): "Y",
        frozenset(("G","C")): "S",
        frozenset(("A","T")): "W",
        frozenset(("G","T")): "K",
        frozenset(("A","C")): "M",
        frozenset(("A","C","G")): "V",
        frozenset(("A","C","T")): "H",
        frozenset(("A","G","T")): "D",
        frozenset(("C","G","T")): "B",
        frozenset(("A","C","G","T")): "N",
    }

    def iupac_for(a, b):
        a = (a or "N").upper()
        b = (b or "N").upper()
        if a == b:
            return a
        return IUPAC.get(frozenset((a, b)), "N")

    ref_seq = ref_fa.fetch(reference=scaffold_name, start=0, end=int(scaffold_len))
    builders = [list(ref_seq) for _ in range(n)]

    rows = rows.sort_values("start", kind="mergesort")

    skipped_overlap = 0
    applied_hom_snv = applied_het_snv = 0
    applied_hom_indel = skipped_het_indel = 0

    for s_idx in range(n):
        seq = builders[s_idx]
        offset = 0
        applied_intervals = []

        for _, r in rows.iterrows():
            ds = str(r.get("diem_genotype", ""))
            if not (ds and ds[0] == "S" and len(ds) == n + 1):
                continue

            ch = ds[s_idx + 1]
            if is_unencodable_char(ch):
                continue

            pair = decode.get(ch, None)
            if pair is None:
                continue

            # Coordinates
            start0_ref = int(r["start"])  # 0-based BED start in reference coordinates
            start0 = start0_ref + offset  # current coordinate after indels applied
            ref_allele = str(r.get("ref", "")) or "N"
            len_ref = len(ref_allele)
            end0 = start0 + len_ref

            # Overlap check (current coordinate space)
            collided = False
            for (s, e) in applied_intervals:
                if not (end0 <= s or e <= start0):
                    collided = True
                    break
            if collided:
                skipped_overlap += 1
                continue

            seq_labels = str(r.get("SeqAlleles", "")).split(",")
            max_rank = len(seq_labels) - 1
            i, j = pair
            if i > max_rank or j > max_rank:
                continue

            a_i = seq_labels[i]
            a_j = seq_labels[j]
            is_hom = (i == j)

            def is_snv_len(x): 
                return isinstance(x, str) and len(x) == 1

            if is_hom:
                allele = a_i or "N"
                if is_snv_len(ref_allele) and is_snv_len(allele):
                    # HOM SNV
                    seq[start0] = allele.upper()
                    applied_hom_snv += 1
                    applied_intervals.append((start0, start0 + 1))
                else:
                    # HOM indel / multibase: replace ref segment with allele
                    del seq[start0:end0]
                    ins = list((allele or "N").upper())
                    for k, ch2 in enumerate(ins):
                        seq.insert(start0 + k, ch2)
                    delta = len(allele) - len_ref
                    offset += delta
                    applied_hom_indel += 1
                    applied_intervals.append((start0, start0 + len(allele)))
            else:
                # HET call
                if is_snv_len(ref_allele) and is_snv_len(a_i) and is_snv_len(a_j):
                    seq[start0] = iupac_for(a_i, a_j)
                    applied_het_snv += 1
                    applied_intervals.append((start0, start0 + 1))
                else:
                    # HET indel / multibase: conservative marker, no length change
                    seq[start0] = "N"
                    skipped_het_indel += 1
                    applied_intervals.append((start0, start0 + 1))

        builders[s_idx] = "".join(seq)

    print(
        f"[recon {scaffold_name}] hom_snv={applied_hom_snv}, het_snv={applied_het_snv}, "
        f"hom_indel={applied_hom_indel}, skipped_overlap={skipped_overlap}, "
        f"het_indel_conservative={skipped_het_indel}"
    )

    return builders


def reconstruct_region(ref_fa, chrom: str, region_start0: int, region_end0: int,
                       rows, sample_names):
    """
    Reconstruct only a reference slice [region_start0, region_end0) for each sample.

    Conservative rules at region boundaries:
      - only apply variants whose BED start is within region
      - skip edits that would extend beyond current region string after offsets
    """
    import pandas as pd

    # filter rows to those starting inside region
    rows = rows[(rows["start"] >= region_start0) & (rows["start"] < region_end0)].copy()
    if rows.empty:
        seq = ref_fa.fetch(reference=chrom, start=region_start0, end=region_end0)
        return [seq] * len(sample_names)

    # Convert to region-local coordinates so reconstruct_scaffold can work on a "mini-chrom"
    rows["start"] = rows["start"] - region_start0
    rows["end"]   = rows["end"]   - region_start0

    # Make a tiny fake scaffold: we pass a “length” equal to region length and fetch reference inside reconstruct_scaffold
    # To do that cleanly, temporarily monkeypatch fetch by wrapping ref_fa.
    class _SliceRef:
        def __init__(self, fa, chrom, start0, end0):
            self.fa = fa
            self.chrom = chrom
            self.start0 = start0
            self.end0 = end0
        def fetch(self, reference, start, end):
            # reference is ignored; start/end are relative to region
            return self.fa.fetch(reference=self.chrom, start=self.start0 + start, end=self.start0 + end)

    ref_slice = _SliceRef(ref_fa, chrom, region_start0, region_end0)
    region_len = region_end0 - region_start0

    # Use your existing reconstruct_scaffold (the “in line with DIEM decode” version you now have)
    return reconstruct_scaffold(ref_slice, chrom, region_len, rows, sample_names)


def rows_overlapping_regions(by_chr_all, regions):
    """
    regions: list[(chrom, start, end, label)]
    Returns a DataFrame of all rows overlapping any region.
    Assumes rows have 'start' (0-based) and represent a single-site call.
    """
    import pandas as pd

    parts = []
    # group regions by chrom for fewer filters
    reg_by_chr = {}
    for chrom, s, e, _lab in regions:
        reg_by_chr.setdefault(str(chrom), []).append((int(s), int(e)))

    for chrom, spans in reg_by_chr.items():
        df = by_chr_all.get(chrom)
        if df is None or df.empty:
            continue
        # rows are single-site; consider overlap by start within [s,e)
        # (if your rows are 1bp intervals, start is enough)
        s0 = df["start"].to_numpy()
        keep = None
        for s, e in spans:
            m = (s0 >= s) & (s0 < e)
            keep = m if keep is None else (keep | m)
        if keep is not None and keep.any():
            parts.append(df.loc[keep])

    if not parts:
        return pd.DataFrame(columns=["chrom","start","end","diem_genotype","SeqAlleles","ref"])
    return pd.concat(parts, ignore_index=True)


# -------------------- Heterozygosity summary --------------------

def het_summary_for_rows(rows: pd.DataFrame, sample_names: np.ndarray) -> Dict[str, dict]:
    """
    Count heterozygous SNVs / non-SNVs / unknown per sample for a scaffold.
    """
    n = len(sample_names)
    counts = {name: {"het_snv":0, "het_nonsnv":0, "Unencodable":0} for name in sample_names}
    for _, r in rows.iterrows():
        d = r["diem_genotype"]
        if not (isinstance(d, str) and d.startswith("S") and len(d) == n+1):
            continue
        seq_labels = (r.get("SeqAlleles","") or "").split(",")
        is_snv = len(seq_labels) >= 2 and len(seq_labels[0]) == 1 and len(seq_labels[1]) == 1
        for i, name in enumerate(sample_names):
            ch = d[1+i]
            if ch == "1":
                if is_snv:
                    counts[name]["het_snv"] += 1
                else:
                    counts[name]["het_nonsnv"] += 1
            elif ch == "_":
                counts[name]["Unencodable"] += 1
    return counts

# -------------------- Distances --------------------

def _pair_block_hamming(args):
    i0, i1, dstrings, n = args
    num = np.zeros((i1-i0, n), dtype=np.float64)
    den = np.zeros((i1-i0, n), dtype=np.int64)
    for d in dstrings:  # 'S'+n
        row = d[1:]
        for i in range(i0, i1):
            ci = row[i]
            if ci == "_": 
                continue
            for j in range(i, n):
                cj = row[j]
                if cj == "_":
                    continue
                num[i-i0, j] += (ci != cj)  # partial credit naturally by aggregating across sites
                den[i-i0, j] += 1
    return num, den

def _pair_block_compat(args):
    i0, i1, dstrings, n = args
    num = np.zeros((i1-i0, n), dtype=np.float64)
    den = np.zeros((i1-i0, n), dtype=np.int64)
    for d in dstrings:
        row = d[1:]
        for i in range(i0, i1):
            ci = row[i]
            for j in range(i, n):
                cj = row[j]
                # missing data allowed: if either is '_' skip (no evidence)
                if ci == "_" or cj == "_":
                    continue
                den[i-i0, j] += 1
                # compatible (distance 0) iff chars equal; otherwise 1
                num[i-i0, j] += (ci != cj)
    return num, den

def _reduce_symmetric(n: int, parts: List[Tuple[np.ndarray,np.ndarray]]) -> Tuple[np.ndarray,np.ndarray]:
    num = np.zeros((n,n), dtype=np.float64)
    den = np.zeros((n,n), dtype=np.int64)
    # parts order matches block index order; we reconstruct blocks by offsets
    # but we pass offsets alongside parts to be explicit:
    raise_on = False
    # Parts are returned in the same order they were mapped with offsets stored; we can reconstruct using zip
    # However simpler: we pass (num_block, den_block, i0, i1)
    # To keep above helpers simple, build alongside offsets in caller.
    return num, den  # placeholder (not used; we do reduction in the caller for clarity)


def OLDcompute_pairwise_from_dstrings(by_chr_dist, sample_names, mode="hamming", threads=1):
    """
    Multi-allelic Dstring distances.
    mode: "hamming" (unphased, partial credit) or "compat" (binary conflict).
    Returns (distance_matrix, coverage_matrix).
    """
    import numpy as np

    # ---- DIEM decode tables (must match vcf2diem) -------------------------
    diemALPHABET = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRS'
    diemMaxVariants = 10
    diemMaxChars = diemMaxVariants * (diemMaxVariants + 1) // 2
    ABSij = (0, 0, 5, 5, 5, 9, 11, 13, 15, 17)
    MAXij = (0, 0, 0, 0, 3, 6, 10, 15, 21, 28)

    # forward encoder (to build reverse map)
    def enc(i, j):
        if max(i, j) >= diemMaxVariants: return None
        n = i + j + ABSij[abs(i - j)] + MAXij[max(i, j)]
        if 0 <= n < diemMaxChars:
            return diemALPHABET[n]
        return None

    # reverse: char -> (i,j) with i<=j; '_' -> None
    decode = {}
    for j in range(diemMaxVariants):
        for i in range(j + 1):
            ch = enc(i, j)
            if ch is not None:
                decode[ch] = (i, j)
    decode['_'] = None  # missing

    # ---- scoring kernels ---------------------------------------------------
    def hamming_score(a, b):
        # a,b : tuples (i,j) or None
        if a is None or b is None:
            return None  # skip
        (i, j), (k, l) = a, b
        A = (i, j); B = (k, l)
        # same unordered set
        if (i == k and j == l) or (i == l and j == k):
            return 0.0
        # both homo different
        if i == j and k == l and i != k:
            return 1.0
        # one homo vs one het
        if i == j and k != l:
            return 0.5 if (i == k or i == l) else 1.0
        if i != j and k == l:
            return 0.5 if (k == i or k == j) else 1.0
        # both het: overlap?
        overlap = (i == k or i == l or j == k or j == l)
        return 0.5 if overlap else 1.0

    def compat_score(a, b):
        if a is None or b is None:
            return None
        (i, j), (k, l) = a, b
        # both homo different → conflict
        if i == j and k == l and i != k:
            return 1.0
        # otherwise conflict only if disjoint allele sets
        overlap = (i == k or i == l or j == k or j == l)
        return 0.0 if overlap else 1.0

    scorer = hamming_score if mode == "hamming" else compat_score

    # ---- accumulate --------------------------------------------------------
    n = len(sample_names)
    num   = np.zeros((n, n), dtype=np.float64)
    denom = np.zeros((n, n), dtype=np.float64)

    for chrom, rows in by_chr_dist.items():
        if rows is None or rows.empty:
            continue
        # only use well-formed Dstrings ('S' + n chars)
        D = rows['diem_genotype'].astype(str).to_numpy()
        D = [ds for ds in D if ds and ds[0] == 'S' and len(ds) == n + 1]
        if not D:
            continue

        # decode once: per-sample list of allele-pairs per site
        # pairs[sample_index][site_index] = (i,j) or None
        sites = len(D)
        pairs = [[None] * sites for _ in range(n)]
        for s_idx, ds in enumerate(D):
            for i in range(n):
                ch = ds[i + 1]
                pairs[i][s_idx] = decode.get(ch, None)

        # pairwise accumulate
        for i in range(n):
            Ai = pairs[i]
            for j in range(i, n):
                Aj = pairs[j]
                s = 0.0; c = 0.0
                for a, b in zip(Ai, Aj):
                    sc = scorer(a, b)
                    if sc is None:
                        continue
                    s += sc; c += 1.0
                num[i, j]   += s;   num[j, i]   += s
                denom[i, j] += c;   denom[j, i] += c

    with np.errstate(divide='ignore', invalid='ignore'):
        dist = num / denom
    # Diagonal: 0 when defined; NaN when no coverage
    for i in range(n):
        if denom[i, i] > 0:
            dist[i, i] = 0.0
    return dist, denom


def compute_pairwise_from_dstrings(by_chr_dist, sample_names, mode="hamming", threads=1):
    """
    Multi-allelic Dstring distances.
    mode:
      - "hamming": unphased, partial credit
      - "compat":  binary conflict (0 compatible, 1 incompatible), skipping Unencodable

    Returns:
      (distance_matrix, coverage_matrix) where distance is NaN if coverage==0.
    """
    import numpy as np

    decode, is_unencodable_char = build_diem_decoder()

    def hamming_score(a, b):
        if a is None or b is None:
            return None
        (i, j), (k, l) = a, b

        # Same unordered diplotype
        if (i == k and j == l) or (i == l and j == k):
            return 0.0

        # Both homozygous different
        if i == j and k == l and i != k:
            return 1.0

        # Homo vs het: partial if allele shared
        if i == j and k != l:
            return 0.5 if (i == k or i == l) else 1.0
        if i != j and k == l:
            return 0.5 if (k == i or k == j) else 1.0

        # Both het: partial if any allele shared
        overlap = (i == k or i == l or j == k or j == l)
        return 0.5 if overlap else 1.0

    def compat_score(a, b):
        if a is None or b is None:
            return None
        (i, j), (k, l) = a, b

        # Both homozygous different => definite conflict
        if i == j and k == l and i != k:
            return 1.0

        # Otherwise: compatible iff allele sets overlap
        overlap = (i == k or i == l or j == k or j == l)
        return 0.0 if overlap else 1.0

    scorer = hamming_score if mode == "hamming" else compat_score

    n = len(sample_names)
    num   = np.zeros((n, n), dtype=np.float64)
    denom = np.zeros((n, n), dtype=np.float64)

    for chrom, rows in by_chr_dist.items():
        if rows is None or rows.empty:
            continue

        D = rows["diem_genotype"].astype(str).to_numpy()
        D = [ds for ds in D if ds and ds[0] == "S" and len(ds) == n + 1]
        if not D:
            continue

        sites = len(D)

        # pairs[sample][site] = (i,j) or None
        pairs = [[None] * sites for _ in range(n)]
        for site_idx, ds in enumerate(D):
            # ds[1:] is per-sample site-state char
            for s_idx in range(n):
                ch = ds[s_idx + 1]
                if is_unencodable_char(ch):
                    pairs[s_idx][site_idx] = None
                else:
                    pairs[s_idx][site_idx] = decode.get(ch, None)

        # accumulate
        for i in range(n):
            Ai = pairs[i]
            for j in range(i, n):
                Aj = pairs[j]
                s = 0.0
                c = 0.0
                for a, b in zip(Ai, Aj):
                    sc = scorer(a, b)
                    if sc is None:
                        continue
                    s += sc
                    c += 1.0
                num[i, j]   += s
                denom[i, j] += c
                if i != j:
                    num[j, i]   += s
                    denom[j, i] += c

    with np.errstate(divide="ignore", invalid="ignore"):
        dist = num / denom

    # IMPORTANT: do NOT force diagonal to 0 if denom==0; keep NaN
    for i in range(n):
        if denom[i, i] > 0:
            dist[i, i] = 0.0

    return dist, denom



def run_summary(
    df_all: "pd.DataFrame",
    df_dist: "pd.DataFrame",
    chr_names,
    chr_lengths,
    sample_names,
    outdir: str,
    *,
    label: str = "diem2fasta",
    dstring_col: str = "diem_genotype",
    excl_col: str = "exclusion_criterion",
    chrom_col: str = "chrom",
    start_col: str = "start",
    end_col: str = "end",
    seqalleles_col: str = "SeqAlleles",
) -> dict:
    """
    Print a compact QC/diagnostic summary and return the same info as a dict.

    Assumptions:
      - df_all is your union (variants + excludes, with E1 removed if you do that)
      - df_dist is the subset actually used for distance matrices (e.g., E0-only)
      - Dstrings are site-strings 'S' + one char per sample
    """
    import os
    import numpy as np
    import pandas as pd

    n_samples = len(sample_names)
    expected_len = n_samples + 1

    def _is_good_dstring(x) -> bool:
        return isinstance(x, str) and len(x) == expected_len and x.startswith("S")

    # ---- Row counts
    n_all = 0 if df_all is None else int(len(df_all))
    n_dist = 0 if df_dist is None else int(len(df_dist))

    # ---- Exclusion counts (top few)
    excl_counts = None
    if df_all is not None and excl_col in df_all.columns:
        excl_counts = df_all[excl_col].astype("string").fillna("").value_counts().head(10)

    # ---- Dstring shape / quality
    good_all = bad_all = good_dist = bad_dist = 0
    if df_all is not None and dstring_col in df_all.columns:
        ds = df_all[dstring_col].tolist()
        good_all = sum(_is_good_dstring(x) for x in ds)
        bad_all  = len(ds) - good_all
    if df_dist is not None and dstring_col in df_dist.columns:
        ds = df_dist[dstring_col].tolist()
        good_dist = sum(_is_good_dstring(x) for x in ds)
        bad_dist  = len(ds) - good_dist

    # ---- Character composition on DIST rows (only well-formed)
    char_counts = None
    het1_total = unenc_total = 0
    if df_dist is not None and dstring_col in df_dist.columns and good_dist:
        # Concatenate all sample chars (drop leading 'S')
        # This is O(total_chars) but only for summary.
        concat = []
        for x in df_dist[dstring_col].tolist():
            if _is_good_dstring(x):
                concat.append(x[1:])
        all_chars = "".join(concat)
        if all_chars:
            from collections import Counter
            char_counts = Counter(all_chars)
            het1_total = char_counts.get("1", 0)
            unenc_total = char_counts.get("_", 0) + char_counts.get("U", 0)

    # ---- SNV vs non-SNV (best-effort)
    snv_rows = nonsnv_rows = unknown_rows = 0
    if df_dist is not None and seqalleles_col in df_dist.columns:
        sa = df_dist[seqalleles_col].astype("string")
        # SNV: first two alleles exist and are length 1
        # (best effort; empty/malformed -> unknown)
        for s in sa.tolist():
            if not s:
                unknown_rows += 1
                continue
            parts = str(s).split(",")
            if len(parts) >= 2 and len(parts[0]) == 1 and len(parts[1]) == 1:
                snv_rows += 1
            elif len(parts) >= 1:
                nonsnv_rows += 1
            else:
                unknown_rows += 1

    # ---- Chrom coverage
    chrom_row_counts = None
    chroms_present = 0
    if df_dist is not None and chrom_col in df_dist.columns:
        chrom_row_counts = (
            df_dist[chrom_col].astype("string").value_counts()
            if len(df_dist) else pd.Series(dtype="int64")
        )
        chroms_present = int((chrom_row_counts > 0).sum())

    # ---- Basic coordinate sanity
    coord_ok = None
    if df_dist is not None and start_col in df_dist.columns and end_col in df_dist.columns:
        try:
            starts = pd.to_numeric(df_dist[start_col], errors="coerce")
            ends   = pd.to_numeric(df_dist[end_col], errors="coerce")
            coord_ok = float(((starts.notna()) & (ends.notna()) & (ends >= starts)).mean())
        except Exception:
            coord_ok = None

    # ---- Build summary dict
    summary = {
        "label": label,
        "n_samples": n_samples,
        "expected_dstring_len": expected_len,
        "rows_all": n_all,
        "rows_dist": n_dist,
        "good_dstrings_all": good_all,
        "bad_dstrings_all": bad_all,
        "good_dstrings_dist": good_dist,
        "bad_dstrings_dist": bad_dist,
        "snv_rows_dist": snv_rows,
        "nonsnv_rows_dist": nonsnv_rows,
        "unknown_seqalleles_rows_dist": unknown_rows,
        "chroms_in_meta": int(len(chr_names)) if chr_names is not None else None,
        "chroms_with_dist_rows": chroms_present,
        "coord_sanity_fraction": coord_ok,
        "het_char_1_total": int(het1_total),
        "unenc_total_(_or_U)": int(unenc_total),
    }

    # ---- Print
    print(f"[{label}] samples: {n_samples} (expected Dstring len={expected_len})")
    print(f"[{label}] rows: all={n_all:,d}  dist={n_dist:,d}")
    print(f"[{label}] Dstrings: good/all={good_all:,d} bad/all={bad_all:,d} | good/dist={good_dist:,d} bad/dist={bad_dist:,d}")
    if coord_ok is not None:
        print(f"[{label}] coord sanity (end>=start & numeric): {coord_ok:.3f}")
    print(f"[{label}] dist rows by type: SNV={snv_rows:,d} non-SNV={nonsnv_rows:,d} unknown={unknown_rows:,d}")
    print(f"[{label}] chroms: meta={len(chr_names) if chr_names is not None else 'NA'}  with-dist-rows={chroms_present}")

    if excl_counts is not None and len(excl_counts):
        print(f"[{label}] exclusion_criterion (top 10):")
        for k, v in excl_counts.items():
            print(f"  {k or '<blank>'}\t{int(v):,d}")

    if char_counts is not None:
        # Show key chars first
        keys = ["0", "1", "2", "_", "U"]
        shown = {k: char_counts.get(k, 0) for k in keys}
        other = sum(v for k, v in char_counts.items() if k not in shown)
        print(f"[{label}] Dstring char totals (dist, excluding 'S'):")
        print("  " + "  ".join([f"{k}:{shown[k]:,d}" for k in keys]) + f"  other:{other:,d}")

    # ---- Write TSV (optional but nice)
    try:
        os.makedirs(outdir, exist_ok=True)
        out_path = os.path.join(outdir, f"{label}_run_summary.tsv")
        pd.DataFrame([summary]).to_csv(out_path, sep="\t", index=False)
        summary["summary_tsv"] = out_path
    except Exception:
        summary["summary_tsv"] = None

    return summary

# -------------------- Orchestrator --------------------

def run_diem2fasta(meta_path: str,
                   variants_path: str,
                   ref_fasta_path: str,
                   excludes_path: Optional[str] = None,
                   outdir: str = "diem2fasta_out",
                   threads: int = 1,
                   out_fasta_path: Optional[str] = None,
                   regions_bed_path: Optional[str] = None
                   ) -> Dict[str, object]:
    """
    Notebook-friendly runner. Writes per-scaffold FASTAs and TSVs; returns dict.
    """
    os.makedirs(outdir, exist_ok=True)
    regions_fasta_written = None

    # META
    chr_names, chr_lengths, sample_names, _ploidy = load_meta(meta_path, ref_fasta_path, variants_path)
    n = len(sample_names)

    # BEDs (strings) and ensure Dstring column name
    df_v = read_bed_as_strings(variants_path)
    _normalize_bed_columns_inplace(df_v)
    df_x = None
    if excludes_path and os.path.exists(excludes_path):
        df_x = read_bed_as_strings(excludes_path)
        _normalize_bed_columns_inplace(df_x)

    # Normalise a few columns
    for df in (df_v, df_x) if df_x is not None else (df_v,):
        for c in ("start","end"):
            if c in df.columns:
                df[c] = pd.to_numeric(df[c], errors="coerce").astype("Int64").astype("int64")
        if "exclusion_criterion" in df.columns:
            df["exclusion_criterion"] = df["exclusion_criterion"].astype("string").str.strip()

    # Union and keep everything except pure invariants (E1)
    df_all = union_variants(df_v, df_x)
    _normalize_bed_columns_inplace(df_all)

    # >>> INSERT #3: enforce required columns + numeric coords on the union <<<
    required = ['chrom', 'start', 'end', 'diem_genotype']
    missing = [c for c in required if c not in df_all.columns]
    if missing:
        print("df_all columns:", list(df_all.columns))  # quick peek
        raise KeyError(f"Required columns missing after normalization: {missing}")
    df_all['chrom'] = df_all['chrom'].astype(str)
    for c in ('start', 'end'):
        df_all[c] = pd.to_numeric(df_all[c], errors='coerce').astype('Int64').astype('int64')
    # <<< END INSERT #3 >>>
    
    # Assert Dstrings are well-formed (if any row exists)
    if not df_all.empty:
        assert_dstring_shape(df_all, n, where="variants+excludes union")

    # E0-only set for distances (matches your baseline)
    df_dist = df_all[df_all.get("exclusion_criterion","E0") == "E0"].copy()

    summary = run_summary(
        df_all=df_all,
        df_dist=df_dist,
        chr_names=chr_names,
        chr_lengths=chr_lengths,
        sample_names=sample_names,
        outdir=outdir,
        label="diem2fasta",
    )

    # Group by chromosome
    by_chr_all  = rows_by_chrom(df_all,  chr_names)
    by_chr_dist = rows_by_chrom(df_dist, chr_names)

    # Reference
    ref = pysam.FastaFile(ref_fasta_path)

    # --- Regions-only FASTA mode ------------------------------------------
    if regions_bed_path:
        print(f"[mode] regions2fasta: {regions_bed_path}")
        
        regions = read_regions_bed(regions_bed_path)
        if out_fasta_path is None:
            out_fasta_path = os.path.join(outdir, "regions.recon.fasta")
        regions_fasta_written = out_fasta_path

        with open(out_fasta_path, "w", encoding="utf-8") as fh:
            for chrom, rstart, rend, label in regions:
                rows = by_chr_all.get(str(chrom))
                if rows is None or rows.empty:
                    seq = ref.fetch(reference=str(chrom), start=int(rstart), end=int(rend))
                    for sname in sample_names:
                        fh.write(f">{label}|{sname}\n")
                        for i in range(0, len(seq), 60):
                            fh.write(seq[i:i+60] + "\n")
                    continue

                seqs = reconstruct_region(ref, str(chrom), int(rstart), int(rend), rows, sample_names)
                for sname, seq in zip(sample_names, seqs):
                    fh.write(f">{label}|{sname}\n")
                    for i in range(0, len(seq), 60):
                        fh.write(seq[i:i+60] + "\n")

        # Still compute het report + distances as usual (or skip, if you want)
        # Return now so NO per-scaffold FASTAs are written.
        het_df = pd.DataFrame({
            "sample": sample_names,
            "het_snv": 0,
            "het_nonsnv": 0,
            "Unencodable": 0,
        })
        # (Optional) keep your existing het computation if you want it in regions-only mode.

        d_ham, cov_ham   = compute_pairwise_from_dstrings(by_chr_dist, sample_names, mode="hamming", threads=threads)
        d_comp, cov_comp = compute_pairwise_from_dstrings(by_chr_dist, sample_names, mode="compat",  threads=threads)

        pd.DataFrame(d_ham,  index=sample_names, columns=sample_names).to_csv(
            os.path.join(outdir, "distance_unphased_hamming.tsv"), sep="\t", na_rep=""
        )
        pd.DataFrame(cov_ham, index=sample_names, columns=sample_names).to_csv(
            os.path.join(outdir, "distance_unphased_hamming_coverage.tsv"), sep="\t"
        )
        pd.DataFrame(d_comp, index=sample_names, columns=sample_names).to_csv(
            os.path.join(outdir, "distance_compatibility.tsv"), sep="\t", na_rep=""
        )
        pd.DataFrame(cov_comp, index=sample_names, columns=sample_names).to_csv(
            os.path.join(outdir, "distance_compatibility_coverage.tsv"), sep="\t"
        )

        print("[note] regions2fasta mode: wrote ONLY regions FASTA (no per-scaffold FASTAs).")

        # Het report in regions-only mode
        from collections import Counter
        het_totals = {name: Counter() for name in sample_names}
        
        df_reg = rows_overlapping_regions(by_chr_all, regions)
        if not df_reg.empty:
            hs = het_summary_for_rows(df_reg, sample_names)
            for name in sample_names:
                het_totals[name].update(hs[name])
        
        het_df = pd.DataFrame({
            "sample": sample_names,
            "het_snv":     [het_totals[nm].get("het_snv",0) for nm in sample_names],
            "het_nonsnv":  [het_totals[nm].get("het_nonsnv",0) for nm in sample_names],
            "Unencodable": [het_totals[nm].get("Unencodable",0) for nm in sample_names],
        })
        het_df.to_csv(os.path.join(outdir, "regions_heterozygosity_report.tsv"), sep="\t", index=False)

        return {
            "regions_fasta": out_fasta_path,
            "het_report": het_df,
            "distance_hamming": pd.DataFrame(d_ham, index=sample_names, columns=sample_names),
            "coverage_hamming": pd.DataFrame(cov_ham, index=sample_names, columns=sample_names),
            "distance_compat": pd.DataFrame(d_comp, index=sample_names, columns=sample_names),
            "coverage_compat": pd.DataFrame(cov_comp, index=sample_names, columns=sample_names),
        }
    # ----------------------------------------------------------------------


    # FASTAs + heterozygosity
    from collections import Counter
    het_totals = {name: Counter() for name in sample_names}
    fasta_paths = []

    for chrom, clen in zip(chr_names, chr_lengths):
        rows = by_chr_all.get(str(chrom))
        if rows is None or rows.empty:
            # write reference for all samples
            seq = ref.fetch(reference=str(chrom), start=0, end=int(clen))
            outfa = os.path.join(outdir, f"{chrom}.recon.fasta")
            with open(outfa, "w") as fh:
                for sname in sample_names:
                    fh.write(f">{sname}\n")
                    for i in range(0, len(seq), 60):
                        fh.write(seq[i:i+60] + "\n")
            fasta_paths.append(outfa)
            continue

        seqs = reconstruct_scaffold(ref, str(chrom), int(clen), rows, sample_names)
        outfa = os.path.join(outdir, f"{chrom}.recon.fasta")
        with open(outfa, "w") as fh:
            for sname, seq in zip(sample_names, seqs):
                fh.write(f">{sname}\n")
                for i in range(0, len(seq), 60):
                    fh.write(seq[i:i+60] + "\n")
        fasta_paths.append(outfa)

        hs = het_summary_for_rows(rows, sample_names)
        for name in sample_names:
            het_totals[name].update(hs[name])

    het_df = pd.DataFrame({
        "sample": sample_names,
        "het_snv":     [het_totals[nm].get("het_snv",0) for nm in sample_names],
        "het_nonsnv":  [het_totals[nm].get("het_nonsnv",0) for nm in sample_names],
        "Unencodable": [het_totals[nm].get("Unencodable",0) for nm in sample_names],
    })
    het_df.to_csv(os.path.join(outdir, "heterozygosity_report.tsv"), sep="\t", index=False)

    # DEBUG FUCK
    # --- Dstring sanity diagnostics (E0 rows only) ---
    def _is_d_ok(s, n):
        # proper Dstring: a string, starts with 'S', length == n+1
        return isinstance(s, str) and s.startswith('S') and len(s) == (n + 1)

    n_e0 = sum((not df.empty) for df in by_chr_dist.values())
    ok = 0
    bad = 0
    for chrom, df in by_chr_dist.items():
        if df is None or df.empty:
            continue
        col = 'diem_genotype' if 'diem_genotype' in df.columns else None
        if col is None:
            continue
        s = df[col].astype(object)
        mask_ok = s.map(lambda x: _is_d_ok(x, n))
        ok += int(mask_ok.sum())
        bad += int((~mask_ok).sum())

    print(f"[diag] E0 contigs with rows: {n_e0}; good Dstrings: {ok}; bad Dstrings: {bad}")
    if ok == 0:
        print("[diag] No usable site Dstrings were found in the E0 set. "
              "Expect NaNs in distance matrices and zero coverages.")

    #END DEBUG FUCK
    
    # Distances from Dstrings (E0 only) — your two flavors
    d_ham, cov_ham   = compute_pairwise_from_dstrings(by_chr_dist, sample_names, mode="hamming", threads=threads)
    d_comp, cov_comp = compute_pairwise_from_dstrings(by_chr_dist, sample_names, mode="compat",  threads=threads)

    pd.DataFrame(d_ham,  index=sample_names, columns=sample_names).to_csv(
        os.path.join(outdir, "distance_unphased_hamming.tsv"), sep="\t", na_rep=""
    )
    pd.DataFrame(cov_ham, index=sample_names, columns=sample_names).to_csv(
        os.path.join(outdir, "distance_unphased_hamming_coverage.tsv"), sep="\t"
    )
    pd.DataFrame(d_comp, index=sample_names, columns=sample_names).to_csv(
        os.path.join(outdir, "distance_compatibility.tsv"), sep="\t", na_rep=""
    )
    pd.DataFrame(cov_comp, index=sample_names, columns=sample_names).to_csv(
        os.path.join(outdir, "distance_compatibility_coverage.tsv"), sep="\t"
    )

    # Policy note
    print("[note] Heterozygous non-SNV sites were written as 'N' (conservative).")

    return {
        "fasta_paths": fasta_paths,
        "het_report": het_df,
        "distance_hamming": pd.DataFrame(d_ham, index=sample_names, columns=sample_names),
        "coverage_hamming": pd.DataFrame(cov_ham, index=sample_names, columns=sample_names),
        "distance_compat": pd.DataFrame(d_comp, index=sample_names, columns=sample_names),
        "coverage_compat": pd.DataFrame(cov_comp, index=sample_names, columns=sample_names),
        "regions_fasta": regions_fasta_written,
    }

def main():
    if "ipykernel" in sys.modules:
        raise SystemExit("diem2fasta: refusing to run CLI inside ipykernel; call run_diem2fasta(...) instead.")
    import argparse
    ap = argparse.ArgumentParser(description="diem2fasta (clean)")
    ap.add_argument("--meta",     required=True)
    ap.add_argument("--variants", required=True)
    ap.add_argument("--ref",      required=True)
    ap.add_argument("--excludes", default=None)
    ap.add_argument("--outdir",   default="diem2fasta_out")
    ap.add_argument("--threads",  type=int, default=max(1, cpu_count()//2))
    ap.add_argument("--out_fasta", default=None, help="Write a single combined FASTA to this path.")
    ap.add_argument("--regions2fasta", default=None, help="BED of regions to reconstruct to FASTA (0-based). Col4 becomes label.")

    args = ap.parse_args()
    run_diem2fasta(
        meta_path=args.meta,
        variants_path=args.variants,
        ref_fasta_path=args.ref,
        excludes_path=args.excludes,
        outdir=args.outdir,
        threads=args.threads,
        out_fasta_path=args.out_fasta,
        regions_bed_path=args.regions2fasta,
    )
# -------------------- CLI (optional) --------------------
if __name__ == "__main__":
    main()
